# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import ipaddress
import os

import gi

from pyatem.converters.protocol import Converter

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk, GLib, Pango

gi.require_version('Handy', '1')
from gi.repository import Handy

import pyatem.converters.converter as conv


class SetupWindow:
    def __init__(self, application):
        self.application = application

        Handy.init()

        self.window = None
        self.titlebar = None
        self.titleleaflet = None
        self.headerbar_side = None
        self.headerbar = None
        self.headergroup = None
        self.leaflet = None
        self.sidebar = None
        self.content = None
        self.listbox = None
        self.main = None
        self.back = None

        self.sg_sidebar = None
        self.sg_main = None

        self.create_window()

        self.create_pages()
        self.window.show_all()
        Gtk.main()

    def create_window(self):
        self.sg_sidebar = Gtk.SizeGroup()
        self.sg_sidebar.set_mode(Gtk.SizeGroupMode.HORIZONTAL)
        self.sg_main = Gtk.SizeGroup()
        self.sg_main.set_mode(Gtk.SizeGroupMode.HORIZONTAL)

        self.headergroup = Handy.HeaderGroup()

        self.window = Handy.Window()
        self.window.set_default_size(640, 480)
        self.window.set_title('Setup')
        self.window.connect('destroy', self.on_main_window_destroy)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.window.add(box)

        self.titlebar = Gtk.Box()
        self.titleleaflet = Handy.Leaflet()
        self.titlebar.add(self.titleleaflet)

        self.headerbar_side = Handy.HeaderBar()
        self.headerbar_side.set_show_close_button(True)
        self.titleleaflet.add(self.headerbar_side)
        self.titleleaflet.child_set(self.headerbar_side, name="sidebar")
        leaflet_sep = Gtk.Separator()
        leaflet_sep.get_style_context().add_class('sidebar')
        self.titleleaflet.add(leaflet_sep)

        self.headerbar = Handy.HeaderBar()
        self.titleleaflet.add(self.headerbar)
        self.titleleaflet.child_set(self.headerbar, name="content")
        self.headerbar.set_title("Setup")
        self.headerbar.set_show_close_button(True)
        self.headerbar.set_hexpand(True)

        self.headergroup.add_header_bar(self.headerbar_side)
        self.headergroup.add_header_bar(self.headerbar)

        self.back = Gtk.Button.new_from_icon_name("go-previous-symbolic", 1)
        self.back.connect("clicked", self.on_back_clicked)
        self.back.set_visible(False)
        self.back.set_no_show_all(True)
        self.headerbar.pack_start(self.back)

        self.leaflet = Handy.Leaflet()
        self.leaflet.set_transition_type(Handy.LeafletTransitionType.SLIDE)
        self.leaflet.connect("notify::folded", self.on_leaflet_change)
        self.leaflet.connect("notify::visible-child", self.on_leaflet_change)
        self.sidebar = Gtk.Box()
        self.sidebar.set_size_request(200, 0)
        self.content = Gtk.Box()
        self.content.props.hexpand = True
        self.leaflet.add(self.sidebar)
        self.leaflet.child_set(self.sidebar, name="sidebar")
        leaflet_sep = Gtk.Separator()
        leaflet_sep.get_style_context().add_class('sidebar')
        self.leaflet.add(leaflet_sep)
        self.leaflet.add(self.content)
        self.leaflet.child_set(self.content, name="content")
        self.leaflet.set_visible_child_name("sidebar")

        self.sg_sidebar.add_widget(self.headerbar_side)
        self.sg_sidebar.add_widget(self.sidebar)

        self.sg_main.add_widget(self.headerbar)
        self.sg_main.add_widget(self.content)

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.sidebar.pack_start(sw, True, True, 0)
        self.listbox = Gtk.ListBox()
        self.listbox.connect('row-activated', self.on_select_page)
        sw.add(self.listbox)

        ms = Gtk.ScrolledWindow()
        ms.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.content.pack_start(ms, True, True, 0)
        self.main = ms

        box.pack_start(self.titlebar, False, True, 0)
        box.pack_start(self.leaflet, True, True, 0)

        self.actionbar = Gtk.ActionBar()
        self.action_revealer = Gtk.Revealer()
        box.pack_start(self.action_revealer, False, True, 0)
        self.action_revealer.add(self.actionbar)
        label = Gtk.Label(label="You have changed settings that need root permissions to save.", xalign=0.0)
        label.set_line_wrap(True)
        self.actionbar.pack_start(label)
        self.action_button = Gtk.Button.new_with_label("Apply")
        self.action_button.get_style_context().add_class('suggested-action')
        self.actionbar.pack_end(self.action_button)
        # self.action_button.connect('clicked', self.on_save_settings)

    def enumerate_hardware(self):
        classes = []
        for name, cls in conv.__dict__.items():
            if isinstance(cls, type) and name not in ['Field', 'ValueField', 'Converter']:
                classes.append(cls)
        return classes

    def create_pages(self):
        classes = self.enumerate_hardware()

        for device in classes:
            if device.is_plugged_in():
                instance = device()
                instance.connect()
                status = instance.get_status()

                name = instance.get_name()
                if name is None:
                    name = device.NAME.replace('Blackmagic design ', '')
                    name = name.replace('Mini', '').replace('Micro', '').replace('Converter', '')
                    name = name.strip()
                label = Gtk.Label(label=name, xalign=0.0)
                label.set_margin_top(8)
                label.set_margin_left(10)
                label.set_margin_right(10)
                label.set_name('row')

                model = device.NAME.replace('Blackmagic design ', '')
                if status == Converter.STATUS_NEED_POWER:
                    model = 'Needs power'
                model = Gtk.Label(label=model, xalign=0.0)
                model.get_style_context().add_class('dim-label')
                model.set_margin_bottom(8)
                model.set_margin_left(10)
                model.set_margin_right(10)

                box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
                box.add(label)
                box.add(model)

                row = Gtk.ListBoxRow()
                row.set_activatable(status == Converter.STATUS_OK)
                row.add(box)
                row.device = instance
                self.listbox.add(row)
            self.listbox.set_selection_mode(Gtk.SelectionMode.NONE)

        return

    def make_page(self, device):
        for child in self.main:
            self.main.remove(child)
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.set_margin_top(12)
        box.set_margin_bottom(12)
        box.set_margin_left(12)
        box.set_margin_right(12)
        self.main.add(box)

        state = device.get_state()
        sections = {}
        for _, field in state.items():
            if field.section not in sections:
                sections[field.section] = []
            sections[field.section].append(field)

        for section, fields in sections.items():

            label = Gtk.Label(label=section, xalign=0.0)
            label.get_style_context().add_class('heading')
            label.set_margin_bottom(8)
            label.set_margin_top(16)
            box.pack_start(label, False, True, 0)
            frame = Gtk.ListBox()
            frame.set_selection_mode(Gtk.SelectionMode.NONE)
            frame.get_style_context().add_class('content')
            frame.set_margin_bottom(12)
            box.pack_start(frame, False, True, 0)

            for field in fields:
                sbox = Gtk.Box()
                sbox.set_margin_top(8)
                sbox.set_margin_bottom(8)
                sbox.set_margin_left(8)
                sbox.set_margin_right(8)
                lbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
                sbox.pack_start(lbox, True, True, 0)
                frame.add(sbox)
                wbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
                sbox.pack_end(wbox, False, False, 0)

                label = Gtk.Label(label=field.label, xalign=0.0)
                lbox.pack_start(label, False, True, 0)

                value = field.value
                if field.dtype == ipaddress.IPv4Address:
                    value = str(value)

                if field.dtype == bool:
                    widget = Gtk.Switch()
                    widget.set_active(field.value)

                    # Make sure I leak some memory
                    widget.field = field
                    widget.device = device
                    field.widget = widget

                    widget.connect('notify::active', self.on_widget_changed)

                    wbox.pack_start(widget, False, False, 0)
                elif field.ro:
                    widget = Gtk.Label(value)
                    widget.set_xalign(0.0)
                    widget.get_style_context().add_class('dim-label')
                    widget.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
                    wbox.pack_start(widget, False, False, 0)
                elif isinstance(field.mapping, dict):
                    widget = Gtk.ComboBoxText()
                    field.widget = widget
                    widget.device = device
                    widget.field = field

                    widget.set_entry_text_column(0)
                    i = 0
                    for key, display in field.mapping.items():
                        widget.append_text(display[1])
                        if key == value:
                            widget.set_active(i)
                        i += 1
                    widget.connect('changed', self.on_widget_changed)
                    wbox.pack_start(widget, False, False, 0)
                elif field.mapping == "dB":
                    widget = Gtk.SpinButton.new_with_range(-40, 12, 0.1)
                    field.widget = widget
                    widget.field = field
                    widget.device = device
                    widget.set_value(value)
                    widget.connect('value-changed', self.on_widget_changed)
                    wbox.pack_start(widget, False, False, 0)
                elif field.dtype == int:
                    w_min = 1
                    w_max = 255
                    w_step = 1
                    widget = Gtk.SpinButton.new_with_range(w_min, w_max, w_step)
                    field.widget = widget
                    widget.field = field
                    widget.device = device
                    widget.set_value(value)
                    widget.connect('value-changed', self.on_widget_changed)
                    wbox.pack_start(widget, False, False, 0)
                elif field.dtype == str or field.dtype == ipaddress.IPv4Address:
                    widget = Gtk.Entry()
                    field.widget = widget
                    widget.field = field
                    widget.device = device
                    widget.set_text(value)
                    widget.connect('activate', self.on_widget_changed)
                    wbox.pack_start(widget, False, False, 0)
                elif field.dtype == open:
                    widget = Gtk.FileChooserButton(title="Select LUT")
                    field.widget = widget
                    widget.field = field
                    widget.device = device
                    widget.connect('file-set', self.on_widget_changed)
                    wbox.pack_start(widget, False, False, 0)

        self.main.show_all()

    def encode_value(self, field, value):
        if field.dtype == str:
            return value.encode()
        elif field.dtype == int:
            value = int(value)
            return value.to_bytes(length=(8 + (value + (value < 0)).bit_length()), byteorder='little')
        return value

    def on_widget_changed(self, widget, *args):
        field = widget.field
        device = widget.device

        if isinstance(field.mapping, dict):
            display = widget.get_active_text()
            for value, d in field.mapping.items():
                if d[1] == display:
                    break
            else:
                ValueError("Unknown mapping value")
        elif field.dtype == bool:
            value = widget.get_active()
        elif field.dtype == str:
            value = widget.get_text()
        elif field.dtype == int:
            value = widget.get_value()
        elif field.dtype == ipaddress.IPv4Address:
            value = widget.get_text()
            value = ipaddress.IPv4Address(value)
        elif field.dtype == open:
            device.set_lut(field.key, widget.get_filename())
            self.make_page(device)
            return
        else:
            return RuntimeError("Missing codepath")

        device.set_value(field, value)

    def on_select_page(self, widget, row):
        if self.listbox.get_selection_mode() == Gtk.SelectionMode.NONE:
            self.listbox.set_selection_mode(Gtk.SelectionMode.SINGLE)
            self.listbox.select_row(row)

        self.make_page(row.device)

        self.headerbar.set_subtitle(row.device.NAME.replace('Blackmagic design ', ''))
        self.leaflet.set_visible_child_name('content')

        # In folded view unselect the row in the listbox
        # so it's possible to go back to the same page
        if self.leaflet.get_folded():
            self.listbox.unselect_row(row)

    def on_main_window_destroy(self, widget):
        Gtk.main_quit()

    def on_back_clicked(self, widget, *args):
        self.leaflet.set_visible_child_name('sidebar')
        self.headerbar.set_subtitle('')

    def on_leaflet_change(self, *args):
        self.titleleaflet.set_visible_child_name(self.leaflet.get_visible_child_name())
        self.back.set_visible(self.leaflet.get_folded())
        if self.leaflet.get_folded():
            self.headerbar_side.set_title("Setup")
        else:
            self.headerbar_side.set_title("")



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import logging

from gi.repository import Gtk, GObject

from gtk_switcher.template_i18n import TemplateLocale
from pyatem.field import TransitionSettingsField


@TemplateLocale(resource_path='/nl/brixit/switcher/ui/mixeffect.glade')
class MixEffectBlock(Gtk.Grid):
    __gtype_name__ = 'MixEffectBlock'

    dsk_box = Gtk.Template.Child()

    program_bus = Gtk.Template.Child()
    preview_bus = Gtk.Template.Child()
    ftb = Gtk.Template.Child()
    ftb_rate = Gtk.Template.Child()
    auto = Gtk.Template.Child()
    cut = Gtk.Template.Child()
    auto_rate = Gtk.Template.Child()
    prev_trans = Gtk.Template.Child()
    dsks = Gtk.Template.Child()
    focus_dummy = Gtk.Template.Child()

    style_mix = Gtk.Template.Child()
    style_dip = Gtk.Template.Child()
    style_wipe = Gtk.Template.Child()
    style_sting = Gtk.Template.Child()
    style_dve = Gtk.Template.Child()

    next_bkgd = Gtk.Template.Child()
    next_key1 = Gtk.Template.Child()
    next_key2 = Gtk.Template.Child()
    next_key3 = Gtk.Template.Child()
    next_key4 = Gtk.Template.Child()

    onair_key1 = Gtk.Template.Child()
    onair_key2 = Gtk.Template.Child()
    onair_key3 = Gtk.Template.Child()
    onair_key4 = Gtk.Template.Child()

    tbar = Gtk.Template.Child()
    transition_progress = Gtk.Template.Child()
    tbar_adj = Gtk.Template.Child()

    def __init__(self, index):
        super(Gtk.Grid, self).__init__()
        self.init_template()
        self.index = index
        self.mode = None

        self.log = logging.getLogger('MixEffectBlock')

        self.tbar_held = False
        self.last_transition_state = False

        self.rate = {
            'mix': '0:00',
            'dip': '0:00',
            'wipe': '0:00',
            'sting': '0:00',
            'dve': '0:00'
        }

        self.menu = None
        self.routers = []
        self.last_program = None
        self.last_preview = None
        self.last_preview_in_program = False

    def __repr__(self):
        return '<MixEffectBlock index={}>'.format(self.index)

    def set_mode(self, mode):
        self.mode = mode

    def frames_to_time(self, frames):
        # WTF BMD
        if self.mode.rate < 29:
            transition_rate = 25
        elif self.mode.rate < 49:
            transition_rate = 30
        elif self.mode.rate < 59:
            transition_rate = 25
        else:
            transition_rate = 30

        if self.mode is None:
            return '00:00'

        seconds = frames // transition_rate
        frames = frames % transition_rate
        return '{:02d}:{:02d}'.format(int(seconds), int(frames))

    def time_to_frames(self, timestr):
        if self.mode.rate < 29:
            transition_rate = 25
        elif self.mode.rate < 49:
            transition_rate = 30
        elif self.mode.rate < 59:
            transition_rate = 25
        else:
            transition_rate = 30

        if self.mode is None:
            return 0

        part = timestr.split(':')
        if len(part) == 1:
            return int(part[0]) * transition_rate
        elif len(part) == 2:
            return (int(part[0]) * transition_rate) + int(part[1])

    def set_class(self, widget, classname, state):
        if state:
            widget.get_style_context().add_class(classname)
        else:
            widget.get_style_context().remove_class(classname)

    def set_dsk(self, enabled):
        if enabled:
            self.dsk_box.get_style_context().remove_class('hidden')
        else:
            self.dsk_box.get_style_context().add_class('hidden')

    def set_inputs(self, buttons):
        # Clear the existing buttons
        for child in self.program_bus:
            child.destroy()
        for child in self.preview_bus:
            child.destroy()

        for top, row in enumerate(buttons):
            for left, button in enumerate(row):
                if button is None:
                    spacer = Gtk.Box()
                    spacer.set_size_request(4, 4)
                    spacer.source_index = -1
                    pspacer = Gtk.Box()
                    pspacer.set_size_request(4, 4)
                    pspacer.source_index = -1

                    self.program_bus.attach(spacer, left, top, 1, 1)
                    self.preview_bus.attach(pspacer, left, top, 1, 1)
                    continue
                active = button.short_name != ""

                label = Gtk.Label(label=button.short_name)
                link = None
                hwthread = None
                for hwthread, config in self.routers:
                    link = None
                    for output_idx in config['outputs']:
                        output = config['outputs'][output_idx]
                        if output['source'] == button.index:
                            link = output_idx
                    if link is not None:
                        break

                btn = Gtk.Button()
                btn.add(label)
                btn.source_index = button.index
                btn.set_sensitive(active)
                btn.set_size_request(48, 48)
                btn.get_style_context().add_class('bmdbtn')
                btn.router_output = link
                if link is not None:
                    btn.router = hwthread
                    btn.get_style_context().add_class('routable')
                if button.index == self.last_program:
                    btn.get_style_context().add_class('program')
                btn.connect('clicked', self.do_program_input_change)
                btn.connect('button-release-event', self.on_mouse_release)
                self.program_bus.attach(btn, left, top, 1, 1)

                plabel = Gtk.Label(label=button.short_name)
                pbtn = Gtk.Button()
                pbtn.add(plabel)
                pbtn.source_index = button.index
                pbtn.set_sensitive(active)
                pbtn.set_size_request(48, 48)
                pbtn.get_style_context().add_class('bmdbtn')
                pbtn.router_output = link
                if link is not None:
                    pbtn.router = hwthread
                    pbtn.get_style_context().add_class('routable')
                if button.index == self.last_preview:
                    if self.last_preview_in_program:
                        pbtn.get_style_context().add_class('program')
                    else:
                        pbtn.get_style_context().add_class('preview')
                pbtn.connect('clicked', self.do_preview_input_change)
                pbtn.connect('button-release-event', self.on_mouse_release)
                self.preview_bus.attach(pbtn, left, top, 1, 1)

        self.program_bus.show_all()
        self.preview_bus.show_all()

    def on_mouse_release(self, widget, event):
        # Only respond on right click
        if event.button != 3:
            return

        if not hasattr(widget, 'router_output') or widget.router_output is None:
            return

        self.menu = Gtk.Menu()
        for index in widget.router.inputs:
            source = widget.router.inputs[index]
            item = Gtk.MenuItem(source['label'])
            item.router = widget.router
            item.router_output = widget.router_output
            item.router_input = int(index)
            item.connect('activate', self.on_route_change)
            self.menu.append(item)
        self.menu.show_all()
        self.menu.popup(None, None, None, None, 0, Gtk.get_current_event_time())

    def on_route_change(self, widget, *args):
        widget.router.change_route(widget.router_output, widget.router_input)

    def program_input_change(self, data):
        self.last_program = data.source
        for btn in self.program_bus:
            if btn.source_index == data.source:
                btn.get_style_context().add_class('program')
            else:
                btn.get_style_context().remove_class('program')

    def preview_input_change(self, data):
        self.last_preview = data.source
        self.last_preview_in_program = data.in_program
        for btn in self.preview_bus:
            if btn.source_index == data.source:
                if data.in_program:
                    btn.get_style_context().add_class('program')
                else:
                    btn.get_style_context().add_class('preview')
            else:
                btn.get_style_context().remove_class('preview')
                btn.get_style_context().remove_class('program')

    @GObject.Signal(name="program-changed", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int),
                    accumulator=GObject.signal_accumulator_true_handled)
    def program_changed(self, *args):
        pass

    @GObject.Signal(name="preview-changed", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int),
                    accumulator=GObject.signal_accumulator_true_handled)
    def preview_changed(self, *args):
        pass

    def do_program_input_change(self, widget):
        self.focus_dummy.grab_focus()
        self.emit("program-changed", self.index, widget.source_index)

    def do_preview_input_change(self, widget):
        self.focus_dummy.grab_focus()
        self.emit("preview-changed", self.index, widget.source_index)

    @Gtk.Template.Callback()
    def on_rate_focus(self, *args):
        self.emit('rate-focus')

    @Gtk.Template.Callback()
    def on_rate_unfocus(self, *args):
        self.emit('rate-unfocus')

    @GObject.Signal(name="rate-focus")
    def rate_focus(self, *args):
        self.disable_shortcuts = True

    @GObject.Signal(name="rate-unfocus")
    def rate_unfocus(self, *args):
        self.disable_shortcuts = False
        self.focus_dummy.grab_focus()

    @Gtk.Template.Callback()
    def on_context_menu(self, *args):
        self.log.debug("CONTEXT!")

    @Gtk.Template.Callback()
    def on_ftb_clicked(self, *args):
        self.focus_dummy.grab_focus()
        self.emit('ftb-clicked', self.index)

    @GObject.Signal(name="ftb-clicked", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int,),
                    accumulator=GObject.signal_accumulator_true_handled)
    def ftb_clicked(self, *args):
        pass

    def set_ftb_state(self, done, transitioning):
        self.set_class(self.ftb, 'program', done)
        self.set_class(self.ftb, 'active', transitioning)

    @Gtk.Template.Callback()
    def on_tbar_button_press_event(self, *args):
        self.tbar_held = True

    @Gtk.Template.Callback()
    def on_tbar_button_release_event(self, *args):
        self.tbar_held = False

    @GObject.Signal(name="tbar-position-changed", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int),
                    accumulator=GObject.signal_accumulator_true_handled)
    def tbar_position_changed(self, *args):
        pass

    @Gtk.Template.Callback()
    def on_tbar_adj_value_changed(self, widget):
        # Ignore value changes if it's not from the user
        if not self.tbar_held:
            return

        val = widget.get_value()
        if val == 9999.0:
            # Transition done
            widget.set_value(0.0)
            self.tbar.set_inverted(not self.tbar.get_inverted())
            self.transition_progress.set_inverted(self.tbar.get_inverted())
            self.emit("tbar-position-changed", self.index, 0)
        else:
            self.emit("tbar-position-changed", self.index, int(val))

    def set_transition_progress(self, data):
        if data.in_transition:
            self.auto.get_style_context().add_class('program')
        else:
            self.auto.get_style_context().remove_class('program')

        if data.in_transition != self.last_transition_state:
            self.last_transition_state = data.in_transition
            if not data.in_transition:
                # Transition just ended, perform the flip
                self.tbar.set_inverted(not self.tbar.get_inverted())
                self.transition_progress.set_inverted(self.tbar.get_inverted())
                self.tbar_adj.set_value(0.0)

        self.transition_progress.set_fraction(data.position / 9999)
        if not self.tbar_held:
            self.tbar_adj.set_value(data.position)

    @GObject.Signal(name="auto-rate-changed", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, str, int),
                    accumulator=GObject.signal_accumulator_true_handled)
    def auto_rate_changed(self, *args):
        pass

    @Gtk.Template.Callback()
    def on_auto_rate_activate(self, *args):
        style = None

        # Get current transition style
        if self.style_mix.get_style_context().has_class('active'):
            style = 'mix'
        elif self.style_dip.get_style_context().has_class('active'):
            style = 'dip'
        elif self.style_wipe.get_style_context().has_class('active'):
            style = 'wipe'
        elif self.style_sting.get_style_context().has_class('active'):
            style = 'sting'
        elif self.style_dve.get_style_context().has_class('active'):
            style = 'dve'

        # Try to parse the new rate, on failure restore last rate
        try:
            frames = self.time_to_frames(self.auto_rate.get_text())
        except Exception as e:
            if style is not None:
                self.auto_rate.set_text(self.rate[style])
            self.log.error(e)
            return

        # Send new rate
        self.emit('auto-rate-changed', self.index, style, frames)

        # Remove focus from the entry so the keyboard shortcuts start working again
        self.focus_dummy.grab_focus()

    def set_auto_rate(self, style, rate):
        label = self.frames_to_time(rate)
        self.rate[style] = label
        active = False
        if style == 'mix' and self.style_mix.get_style_context().has_class('active'):
            active = True
        elif style == 'dip' and self.style_dip.get_style_context().has_class('active'):
            active = True
        elif style == 'wipe' and self.style_wipe.get_style_context().has_class('active'):
            active = True
        elif style == 'sting' and self.style_sting.get_style_context().has_class('active'):
            active = True
        elif style == 'dve' and self.style_dve.get_style_context().has_class('active'):
            active = True

        if active:
            self.auto_rate.set_text(label)

    def set_transition_settings(self, data):
        self.set_class(self.style_mix, 'active', data.style == TransitionSettingsField.STYLE_MIX)
        self.set_class(self.style_dip, 'active', data.style == TransitionSettingsField.STYLE_DIP)
        self.set_class(self.style_wipe, 'active', data.style == TransitionSettingsField.STYLE_WIPE)
        self.set_class(self.style_sting, 'active', data.style == TransitionSettingsField.STYLE_STING)
        self.set_class(self.style_dve, 'active', data.style == TransitionSettingsField.STYLE_DVE)

        self.set_class(self.next_bkgd, 'active', data.next_transition_bkgd)
        self.set_class(self.next_key1, 'active', data.next_transition_key1)
        self.set_class(self.next_key2, 'active', data.next_transition_key2)
        self.set_class(self.next_key3, 'active', data.next_transition_key3)
        self.set_class(self.next_key4, 'active', data.next_transition_key4)

        if data.style == TransitionSettingsField.STYLE_MIX:
            self.auto_rate.set_text(self.rate['mix'])
        elif data.style == TransitionSettingsField.STYLE_DIP:
            self.auto_rate.set_text(self.rate['dip'])
        elif data.style == TransitionSettingsField.STYLE_WIPE:
            self.auto_rate.set_text(self.rate['wipe'])
        elif data.style == TransitionSettingsField.STYLE_STING:
            self.auto_rate.set_text(self.rate['sting'])
        elif data.style == TransitionSettingsField.STYLE_DVE:
            self.auto_rate.set_text(self.rate['dve'])

    @GObject.Signal(name="auto-clicked", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int,),
                    accumulator=GObject.signal_accumulator_true_handled)
    def auto_clicked(self, *args):
        pass

    @Gtk.Template.Callback()
    def on_auto_clicked(self, *args):
        self.emit('auto-clicked', self.index)
        self.focus_dummy.grab_focus()

    @GObject.Signal(name="cut-clicked", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int,),
                    accumulator=GObject.signal_accumulator_true_handled)
    def cut_clicked(self, *args):
        pass

    @Gtk.Template.Callback()
    def on_cut_clicked(self, *args):
        self.emit('cut-clicked', self.index)
        self.focus_dummy.grab_focus()

    @GObject.Signal(name="preview-transition-clicked", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, bool),
                    accumulator=GObject.signal_accumulator_true_handled)
    def prev_trans_clicked(self, *args):
        pass

    @Gtk.Template.Callback()
    def on_prev_trans_clicked(self, widget):
        current = widget.get_style_context().has_class('program')
        self.emit('preview-transition-clicked', self.index, not current)
        self.focus_dummy.grab_focus()

    def set_preview_transition(self, enabled):
        self.set_class(self.prev_trans, 'program', enabled)

    @GObject.Signal(name="style-changed", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, str),
                    accumulator=GObject.signal_accumulator_true_handled)
    def style_changed(self, *args):
        pass

    @Gtk.Template.Callback()
    def on_style_mix_clicked(self, *args):
        self.emit('style-changed', self.index, 'mix')
        self.focus_dummy.grab_focus()

    @Gtk.Template.Callback()
    def on_style_dip_clicked(self, *args):
        self.emit('style-changed', self.index, 'dip')
        self.focus_dummy.grab_focus()

    @Gtk.Template.Callback()
    def on_style_wipe_clicked(self, *args):
        self.emit('style-changed', self.index, 'wipe')
        self.focus_dummy.grab_focus()

    @Gtk.Template.Callback()
    def on_style_sting_clicked(self, *args):
        self.emit('style-changed', self.index, 'sting')
        self.focus_dummy.grab_focus()

    @Gtk.Template.Callback()
    def on_style_dve_clicked(self, *args):
        self.emit('style-changed', self.index, 'dve')
        self.focus_dummy.grab_focus()

    def set_key_on_air(self, data):
        if data.keyer == 0:
            self.set_class(self.onair_key1, 'program', data.enabled)
        elif data.keyer == 1:
            self.set_class(self.onair_key2, 'program', data.enabled)
        elif data.keyer == 2:
            self.set_class(self.onair_key3, 'program', data.enabled)
        elif data.keyer == 4:
            self.set_class(self.onair_key4, 'program', data.enabled)

    @GObject.Signal(name="onair-clicked", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int, bool),
                    accumulator=GObject.signal_accumulator_true_handled)
    def onair_clicked(self, *args):
        pass

    @Gtk.Template.Callback()
    def on_onair_key1_clicked(self, widget):
        enabled = not widget.get_style_context().has_class('program')
        self.emit('onair-clicked', self.index, 0, enabled)
        self.focus_dummy.grab_focus()

    @Gtk.Template.Callback()
    def on_onair_key2_clicked(self, widget):
        enabled = not widget.get_style_context().has_class('program')
        self.emit('onair-clicked', self.index, 1, enabled)
        self.focus_dummy.grab_focus()

    @Gtk.Template.Callback()
    def on_onair_key3_clicked(self, widget):
        enabled = not widget.get_style_context().has_class('program')
        self.emit('onair-clicked', self.index, 2, enabled)
        self.focus_dummy.grab_focus()

    @Gtk.Template.Callback()
    def on_onair_key4_clicked(self, widget):
        enabled = not widget.get_style_context().has_class('program')
        self.emit('onair-clicked', self.index, 3, enabled)
        self.focus_dummy.grab_focus()

    @GObject.Signal(name="next-clicked", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int),
                    accumulator=GObject.signal_accumulator_true_handled)
    def next_clicked(self, *args):
        pass

    @Gtk.Template.Callback()
    def on_next_clicked(self, widget):
        if widget.get_style_context().has_class('active'):
            widget.get_style_context().remove_class('active')
        else:
            widget.get_style_context().add_class('active')

        current = 0
        if self.next_bkgd.get_style_context().has_class('active'):
            current |= (1 << 0)
        if self.next_key1.get_style_context().has_class('active'):
            current |= (1 << 1)
        if self.next_key2.get_style_context().has_class('active'):
            current |= (1 << 2)
        if self.next_key3.get_style_context().has_class('active'):
            current |= (1 << 3)
        if self.next_key4.get_style_context().has_class('active'):
            current |= (1 << 4)

        self.emit('next-clicked', self.index, current)
        self.focus_dummy.grab_focus()

    def set_topology(self, data):
        for child in self.dsks:
            child.destroy()

        for i in range(0, data.downstream_keyers):
            tie_label = Gtk.Label(label=_("TIE"))
            tie = Gtk.Button()
            tie.add(tie_label)
            tie.dsk_tie = i
            tie.set_size_request(48, 48)
            tie.get_style_context().add_class('bmdbtn')
            tie.connect('clicked', self.do_dsk_tie_clicked)
            self.dsks.attach(tie, i, 0, 1, 1)

            rate_label = Gtk.Label(label=_("rate"))
            rate_label.get_style_context().add_class('dim-label')
            rate_label.get_style_context().add_class('rate')
            rate_entry = Gtk.Entry()
            rate_entry.get_style_context().add_class('rate')
            rate_entry.set_size_request(48, 0)
            rate_entry.set_width_chars(5)
            rate_entry.set_max_width_chars(5)
            rate_entry.set_alignment(0.5)
            rate_entry.connect('focus-in-event', self.on_rate_focus)
            rate_entry.connect('focus-out-event', self.on_rate_unfocus)
            rate_entry.connect('activate', self.do_dsk_rate_activate)
            rate_entry.dsk_rate = i
            rate_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            rate_box.pack_start(rate_label, 1, 1, 1)
            rate_box.dsk_rate_box = i
            rate_box.pack_start(rate_entry, 1, 1, 1)
            self.dsks.attach(rate_box, i, 1, 1, 1)

            air_label = Gtk.Label(label=_("ON\nAIR"))
            air = Gtk.Button()
            air.add(air_label)
            air.dsk_onair = i
            air.set_size_request(48, 48)
            air.get_style_context().add_class('bmdbtn')
            air.connect('clicked', self.do_dsk_onair_clicked)
            self.dsks.attach(air, i, 2, 1, 1)

            auto_label = Gtk.Label(label=_("AUTO"))
            auto = Gtk.Button()
            auto.add(auto_label)
            auto.dsk_auto = i
            auto.set_size_request(48, 48)
            auto.get_style_context().add_class('bmdbtn')
            auto.connect('clicked', self.do_dsk_auto_clicked)
            self.dsks.attach(auto, i, 3, 1, 1)
        self.dsks.show_all()

    def set_dsk(self, data):
        for child in self.dsks:
            if hasattr(child, 'dsk_tie') and child.dsk_tie == data.index:
                self.set_class(child, 'active', data.tie)
            if hasattr(child, 'dsk_rate_box') and child.dsk_rate_box == data.index:
                for bc in child:
                    if hasattr(bc, 'dsk_rate'):
                        label = self.frames_to_time(data.rate)
                        bc.set_text(label)

    def set_dsk_state(self, data):
        for child in self.dsks:
            if hasattr(child, 'dsk_onair') and child.dsk_onair == data.index:
                self.set_class(child, 'program', data.on_air)
            if hasattr(child, 'dsk_auto') and child.dsk_auto == data.index:
                self.set_class(child, 'active', data.is_autotransitioning)

    @GObject.Signal(name="dsk-tie", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int, bool),
                    accumulator=GObject.signal_accumulator_true_handled)
    def dsk_tie_clicked(self, *args):
        pass

    def do_dsk_tie_clicked(self, widget, *args):
        state = not widget.get_style_context().has_class('active')
        self.emit('dsk-tie', self.index, widget.dsk_tie, state)
        self.focus_dummy.grab_focus()

    @GObject.Signal(name="dsk-onair", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int, bool),
                    accumulator=GObject.signal_accumulator_true_handled)
    def dsk_onair_clicked(self, *args):
        pass

    def do_dsk_onair_clicked(self, widget):
        state = not widget.get_style_context().has_class('program')
        self.focus_dummy.grab_focus()
        self.emit('dsk-onair', self.index, widget.dsk_onair, state)

    @GObject.Signal(name="dsk-auto", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int),
                    accumulator=GObject.signal_accumulator_true_handled)
    def dsk_auto_clicked(self, *args):
        pass

    def do_dsk_auto_clicked(self, widget):
        self.focus_dummy.grab_focus()
        self.emit('dsk-auto', self.index, widget.dsk_auto)

    @GObject.Signal(name="dsk-rate", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int, int),
                    accumulator=GObject.signal_accumulator_true_handled)
    def dsk_rate_changed(self, *args):
        pass

    def do_dsk_rate_activate(self, widget, *args):
        # Try to parse the new rate
        try:
            frames = self.time_to_frames(widget.get_text())
        except Exception as e:
            self.log.error(e)
            return
        self.emit('dsk-rate', self.index, widget.dsk_rate, frames)
        self.focus_dummy.grab_focus()

    @GObject.Signal(name="ftb-rate", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(int, int),
                    accumulator=GObject.signal_accumulator_true_handled)
    def ftb_rate_changed(self, *args):
        pass

    @Gtk.Template.Callback()
    def on_ftb_rate_activate(self, widget, *args):
        # Try to parse the new rate
        try:
            frames = self.time_to_frames(widget.get_text())
        except Exception as e:
            self.log.error(e)
            return
        self.emit('ftb-rate', self.index, frames)
        self.focus_dummy.grab_focus()

    def set_ftb_rate(self, frames):
        label = self.frames_to_time(frames)
        self.ftb_rate.set_text(label)

    def set_config(self, data):
        self.next_key1.set_sensitive(data.keyers > 0)
        self.next_key2.set_sensitive(data.keyers > 1)
        self.next_key3.set_sensitive(data.keyers > 2)
        self.next_key4.set_sensitive(data.keyers > 3)

        self.onair_key1.set_sensitive(data.keyers > 0)
        self.onair_key2.set_sensitive(data.keyers > 1)
        self.onair_key3.set_sensitive(data.keyers > 2)
        self.onair_key4.set_sensitive(data.keyers > 3)

    def add_router(self, hwthread, config):
        self.routers.append((hwthread, config))
        for btn in list(self.program_bus) + list(self.preview_bus):
            link = None
            for output_idx in config['outputs']:
                output = config['outputs'][output_idx]
                if output['source'] == btn.source_index:
                    link = output_idx
            btn.router_output = link
            if link is not None:
                btn.router = hwthread
                btn.get_style_context().add_class('routable')
            else:
                btn.get_style_context().remove_class('routable')



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import time
from urllib.parse import urlparse, quote

import gi

from gtk_switcher.eqcurve import EqCurve
from gtk_switcher.preferences import PreferencesWindow
from pyatem import locate
from pyatem.command import FairlightStripPropertiesCommand
from pyatem.field import FairlightStripPropertiesField

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, GdkPixbuf

gi.require_version('Handy', '1')
from gi.repository import Handy


class PresetWindow:
    def __init__(self, parent, provider):
        self.provider = provider
        self.model_changing = True
        builder = Gtk.Builder()
        builder.set_translation_domain("openswitcher")
        builder.add_from_resource('/nl/brixit/switcher/ui/preset.glade')
        builder.connect_signals(self)

        self.window = builder.get_object("dialog")
        self.window.set_transient_for(parent)
        self.window.set_modal(True)

        self.name = builder.get_object('name')

    def run(self):
        return self.window.run()

    def get_name(self):
        return self.name.get_text()

    def destroy(self):
        self.window.destroy()



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import os
from unittest import TestCase
import gzip

import pyatem.media
from PIL import Image


class Test(TestCase):

    def setUp(self):
        fixtures_dir = os.environ.get('TEST_FIXTURES', os.path.join(os.path.dirname(__file__), 'fixtures'))
        fixture = self._encode_test(os.path.join(fixtures_dir, 'ramps.png'))
        self.encoded = pyatem.media.rgb_to_atem(fixture, 1920, 1080)

        with gzip.open(os.path.join(fixtures_dir, 'ramps-atemsc.data.gz'), 'rb') as handle:
            self.reference = handle.read()

    def _encode_test(self, path):
        """
        Load a test fixture image with pillow and run through the atem frame encoder without compression
        """
        im = Image.open(path)
        pixels = im.getdata()
        flat = [item for sublist in pixels for item in sublist]
        return bytes(flat)

    def _decompose(self, raw):
        """
        Converts 8 packed pixels to a list of values
        """
        a = int.from_bytes(raw[0:4], byteorder='big', signed=False)
        b = int.from_bytes(raw[4:8], byteorder='big', signed=False)

        a1 = a >> 20
        cb = (a >> 10) & 0b1111111111
        y1 = a & 0b1111111111
        a2 = b >> 20
        cr = (b >> 10) & 0b1111111111
        y2 = b & 0b1111111111
        return {
            "y1": y1,
            "y2": y2,
            "cr": cr,
            "cb": cb,
            "a1": a1,
            "a2": a2,
        }

    def _test_primary(self, name, rgb, expect):
        sequence = rgb + b'\xff' + rgb + b'\xff'
        encoded = pyatem.media.rgb_to_atem(sequence, 2, 1)
        dec = self._decompose(encoded)
        self.assertAlmostEqual(expect[0], dec['y1'], msg=f'{name} Y', delta=2)
        self.assertAlmostEqual(expect[1], dec['cb'], msg=f'{name} Cb', delta=2)
        self.assertAlmostEqual(expect[2], dec['cr'], msg=f'{name} Cr', delta=2)

    def _assertClose(self, a, b, msg):
        for key in a:
            self.assertAlmostEqual(a[key], b[key], msg=f'{msg} [{key}]', delta=5)

    def _compare_row(self, test, reference, row):
        stride = 1920 * 4
        offset = row * stride

        test = test[offset:offset + stride]
        reference = reference[offset:offset + stride]
        for i in range(0, stride, 8):
            testpixels = test[0:8]
            test = test[8:]
            referencepixels = reference[0:8]
            reference = reference[8:]

            r = self._decompose(referencepixels)
            t = self._decompose(testpixels)
            self._assertClose(r, t, f'chunk {i}')

    def test_greyramp(self):
        self._compare_row(self.encoded, self.reference, 1042)

    def test_redramp(self):
        self._compare_row(self.encoded, self.reference, 963)

    def test_greenramp(self):
        self._compare_row(self.encoded, self.reference, 885)

    def test_blueramp(self):
        self._compare_row(self.encoded, self.reference, 800)

    def test_hues(self):
        self._compare_row(self.encoded, self.reference, 0)

    def test_primaries(self):
        self._test_primary('red', b'\xff\x00\x00', (63 * 4, 102 * 4, 240 * 4))
        self._test_primary('green', b'\x00\xFF\x00', (173 * 4, 42 * 4, 26 * 4))
        self._test_primary('blue', b'\x00\x00\xFF', (32 * 4, 240 * 4, 118 * 4))
        self._test_primary('white', b'\xFF\xFF\xFF', (235 * 4, 128 * 4, 128 * 4))
        self._test_primary('black', b'\x00\x00\x00', (16 * 4, 128 * 4, 128 * 4))
        self._test_primary('yellow', b'\xFF\xFF\x00', (219 * 4, 16 * 4, 138 * 4))
        self._test_primary('cyan', b'\x00\xFF\xFF', (188 * 4, 154 * 4, 16 * 4))
        self._test_primary('magenta', b'\xFF\x00\xFF', (78 * 4, 214 * 4, 230 * 4))



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import struct
import pyatem.mediaconvert as mc


def atem_to_image(data, width, height):
    """Decompress and decode an atem frame to RGBA8888"""
    data = rle_decode(data)
    data = mc.atem_to_rgb(data, width, height)
    return data


def image_to_atem(data, width, height):
    """Convert a frame to atem format and compress it"""
    data = mc.rgb_to_atem(data, width, height)
    return rle_encode(data)


def atem_to_rgb(data, width, height):
    """Wrapper for the native function"""
    return mc.atem_to_rgb(data, width, height)


def rgb_to_atem(data, width, height, premultiply=False):
    """Wrapper for the native function"""
    return mc.rgb_to_atem(data, width, height, premultiply)


def rle_encode_slow(data):
    """
    See rle_decode for format description
    :param data:
    :return:
    """
    result = bytearray()
    lastblock = None
    lastcount = 0
    offset = 0
    while True:
        block = data[offset:offset + 8]
        offset += 8
        if len(block) < 8:
            break
        if block != lastblock:
            if lastcount > 2:
                result += b'\xfe\xfe\xfe\xfe\xfe\xfe\xfe\xfe'
                result += struct.pack('>Q', lastcount)
                result += lastblock
            elif lastcount > 0:
                result += lastblock * lastcount

            result += block
            lastblock = block
            lastcount = 0
            continue
        lastcount += 1
        lastblock = block
    if lastcount > 0:
        result += b'\xfe\xfe\xfe\xfe\xfe\xfe\xfe\xfe'
        result += struct.pack('>Q', lastcount)
        result += lastblock
    return result


def rle_encode(data):
    return mc.rle_encode(data)


def rle_decode(data):
    """
    ATEM frames are compressed with a custom RLE encoding. Data in the frame is grouped in 8 byte chunks since
    that is exactly 2 pixels in the 10-bit YCbCr 4:2:2 data. Most of the data is sent without compression but
    if a 8 byte chunk is fully 0xfe then the following chunk is RLE compressed.

    An RLE compressed part is an 64 bit integer setting the repeat count following the 8-byte block of data
    to be repeated. This seems mainly useful to compress solid colors.

    :param data:
    :return:
    """
    result = bytearray()
    offset = 0
    in_size = len(data)

    while offset < in_size:
        block = data[offset:offset + 8]
        if data[offset] == 0xfe \
                and data[offset + 1] == 0xfe \
                and data[offset + 2] == 0xfe \
                and data[offset + 3] == 0xfe \
                and data[offset + 4] == 0xfe \
                and data[offset + 5] == 0xfe \
                and data[offset + 6] == 0xfe \
                and data[offset + 7] == 0xfe:
            # Got an RLE block
            offset += 8
            count, = struct.unpack_from('>Q', data, offset)
            offset += 8
            block = data[offset:offset + 8]
            result += block * count
            offset += 8
        else:
            # Raw data block
            result += block
            offset += 8
    return result


if __name__ == '__main__':
    with open('/workspace/test.bin', 'wb') as handle:
        handle.write(
            rle_decode(b'\xfe\xfe\xfe\xfe\xfe\xfe\xfe\xfe\x00\x00\x00\x00\x00\x0f\xd2\x00:\x96d\xfa:\x9e\xfc\xfa'))
    exit(0)

    with open('/workspace/usb-0-0.bin', 'rb') as handle:
        raw = handle.read()
    atem_to_image(raw, 1920, 1080)



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import select
import socket
import struct
import logging
import time
from queue import Queue, Empty
import collections
from urllib.parse import urlparse
import threading

import usb.core
import usb.util

from pyatem.socketqueue import SocketQueue
from pyatem.transfer import TransferQueueFlushed, TransferTask


class ConnectionReady:
    def __init__(self):
        pass


class Packet:
    STRUCT_HEADER = struct.Struct('>HHH 2x HH')
    STRUCT_USB = struct.Struct('<I')

    def __init__(self):
        self.flags = 0
        self.length = 0
        self.session = 0
        self.sequence_number = 0
        self.acknowledgement_number = 0
        self.remote_sequence_number = 0
        self.data = None
        self.debug = False
        self.original = None
        self.label = None
        self.last_packet_time = None

    @classmethod
    def from_bytes(cls, packet):
        res = cls()
        res.original = packet
        fields = cls.STRUCT_HEADER.unpack_from(packet)
        res.length = fields[0] & ~(0x1f << 11)
        res.flags = (fields[0] & (0x1f << 11)) >> 11

        if res.length != len(packet):
            raise ValueError(
                "Incomplete or corrupt packet received, {} in header but data length is {}".format(
                    res.length, len(packet)))

        res.session = fields[1]
        res.acknowledgement_number = fields[2]
        res.remote_sequence_number = fields[3]
        res.sequence_number = fields[4]
        res.data = packet[12:]
        return res

    def to_bytes(self):
        header_len = 12
        data_len = len(self.data) if self.data is not None else 0
        packet_len = header_len + data_len
        result = self.STRUCT_HEADER.pack(
            packet_len + (self.flags << 11),
            self.session,
            self.acknowledgement_number,
            self.remote_sequence_number,
            self.sequence_number)

        if self.data:
            result += bytes(self.data)

        return result

    def to_usb(self):
        data_len = len(self.data) if self.data is not None else 0
        result = self.STRUCT_USB.pack(data_len)
        if self.data:
            result += bytes(self.data)
        return result

    def __repr__(self):
        flags = ''
        extra = ''
        if self.flags & UdpProtocol.FLAG_RELIABLE:
            flags += ' RELIABLE'
        if self.flags & UdpProtocol.FLAG_SYN:
            flags += ' SYN'
        if self.flags & UdpProtocol.FLAG_RETRANSMISSION:
            flags += ' RETRANSMISSION'
        if self.flags & UdpProtocol.FLAG_REQUEST_RETRANSMISSION:
            flags += ' REQ-RETRANSMISSION'
            extra = ' req={}'.format(self.remote_sequence_number)
        if self.flags & UdpProtocol.FLAG_ACK:
            flags += ' ACK'
            extra = ' ack={}'.format(self.acknowledgement_number)
        flags = flags.strip()
        data_len = len(self.data) if self.data is not None else 0
        label = ''
        if self.label:
            label = ' ' + self.label
        return '<Packet flags={} data={} sequence={}{}{}>'.format(flags, data_len, self.sequence_number, extra, label)

    def get_flags(self):
        flags = [hex(self.flags), len(self.data)]
        if self.flags & 0x01:
            flags.append('AckRequest')
        if self.flags & 0x02:
            flags.append('Hello')
        if self.flags & 0x04:
            flags.append('Resend')
        if self.flags & 0x08:
            flags.append('Undefined')
        if self.flags & 0x10:
            flags.append('Ack')
        return flags


class BaseProtocol:
    def __init__(self):
        self.send_queue = collections.deque(maxlen=1024)
        self.queue_enabled = False
        self.queue_callback = None
        self.mark_next_connected = False
        self.batch_size = 1
        self.batch_delay = 0

    def _send_packet(self, packet):
        raise NotImplementedError()

    def queue_packet(self, packet):
        self.send_queue.append(packet)

    def queue_trigger(self):
        if len(self.send_queue) > 0:
            self.queue_enabled = True
            for i in range(0, min(len(self.send_queue), self.batch_size)):
                p = self.send_queue.popleft()
                self._send_packet(p)
                if self.queue_callback is not None:
                    self.queue_callback(len(self.send_queue), len(p.data) - 4)
            time.sleep(self.batch_delay)
        elif self.queue_enabled:
            self.queue_enabled = False
            return True
        return False

    def get_link_quality(self):
        return 100


class UdpProtocol(BaseProtocol):
    STATE_CLOSED = 0
    STATE_SYN_SENT = 1
    STATE_SYN_RECEIVED = 2
    STATE_ESTABLISHED = 3

    FLAG_RELIABLE = 1
    FLAG_SYN = 2
    FLAG_RETRANSMISSION = 4
    FLAG_REQUEST_RETRANSMISSION = 8
    FLAG_ACK = 16

    def __init__(self, ip, port=9910):
        super().__init__()
        self.ip = ip
        self.port = port

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.settimeout(5)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1024 * 1024 * 16)

        self.thread = threading.Thread(None, self._udp_thread, "atem-udp", daemon=True)

        self.local_sequence_number = 0
        self.local_ack_number = 0
        self.remote_sequence_number = 0
        self.remote_ack_numbe = 0

        self.state = UdpProtocol.STATE_CLOSED
        self.session_id = 0x1337

        self.enable_ack = False
        self.had_traffic = False

        self.received_packets = collections.deque(maxlen=1024)
        self.retransmission_buffer = {}

        self.thread_queue = SocketQueue()
        self.thread_recv_queue = Queue()

        self.batch_size = 5
        self.batch_delay = 0.003

        self.log = logging.getLogger('UdpTransport')
        self.packet_sucess = 0
        self.packet_errors = 0

    def _udp_thread(self):
        while True:
            readable, _, _ = select.select([self.sock, self.thread_queue], [], [])
            for queue in readable:
                if queue is self.sock:
                    packet = self._receive_packet_low()
                    if packet is not None:
                        self.thread_recv_queue.put(packet)
                elif queue is self.thread_queue:
                    try:
                        self._send_packet_low(queue.get())
                    except OSError as e:
                        self.log.error(e)
                        # Queue a None to signal the socket died
                        self.thread_recv_queue.put(None)
                        return
                else:
                    self.thread_recv_queue.put(None)
                    RuntimeError("Unexpected result from select()")

    def get_link_quality(self):
        return 100 - (self.packet_errors / self.packet_sucess * 100)

    def _send_packet(self, packet):
        self.thread_queue.put(packet)
        self.packet_sucess += 1

    def _send_packet_low(self, packet):
        packet.session = self.session_id
        if not packet.flags & UdpProtocol.FLAG_ACK:
            if self.local_sequence_number == -1:
                self.local_sequence_number = 0
            packet.sequence_number = (self.local_sequence_number + 1) % 2 ** 16
        raw = packet.to_bytes()
        self.sock.sendto(raw, (self.ip, self.port))
        self.log.debug('> {}'.format(packet))
        if packet.debug:
            # hexdump(raw)
            pass
        if packet.flags & (UdpProtocol.FLAG_SYN | UdpProtocol.FLAG_ACK) == 0:
            self.local_sequence_number = (self.local_sequence_number + 1) % 2 ** 16
        self.retransmission_buffer[self.local_sequence_number] = packet

        if packet.label == "_handshake":
            # Clear temporary session id, use the session id received in the first packet from the remote
            self.session_id = None

    def _receive_packet(self):
        return self.thread_recv_queue.get()

    def _receive_packet_low(self):
        try:
            data, address = self.sock.recvfrom(2048)
        except socket.timeout:
            # No longer receiving data from the hardware, reset the state of the connection and re-init
            self.state = UdpProtocol.STATE_CLOSED
            self.connect()
            return
        packet = Packet.from_bytes(data)

        if packet.flags & UdpProtocol.FLAG_RETRANSMISSION:
            if len(data) > 12:
                self.log.error("retransmission detected")
                self.packet_errors += 1
                if self.get_link_quality() < 80:
                    self.log.error(f"Connection quality bad ({self.get_link_quality():.1f}%)")
            else:
                self.log.warning("retransmission of PING detected, non-fatal")
        else:
            self.packet_sucess += 1

        if packet.flags & UdpProtocol.FLAG_REQUEST_RETRANSMISSION:
            self.log.error("retransmission requested")
            self.packet_errors += 1
            # hexdump(data)

        new_sequence_number = packet.sequence_number
        self.remote_sequence_number = new_sequence_number

        if self.session_id is None:
            self.session_id = packet.session

        is_retransmissions = packet.flags & UdpProtocol.FLAG_RETRANSMISSION
        if is_retransmissions:
            if self.remote_sequence_number in self.received_packets:
                return True

        self.received_packets.append(self.remote_sequence_number)

        if (packet.flags & UdpProtocol.FLAG_RELIABLE and self.enable_ack) or \
                (not self.enable_ack and UdpProtocol.FLAG_ACK and len(packet.data) == 0):
            self.enable_ack = True
            # ACK this
            ack = Packet()
            ack.flags = UdpProtocol.FLAG_ACK
            ack.acknowledgement_number = self.remote_sequence_number
            ack.remote_sequence_number = 0x61
            self._send_packet(ack)

        return packet

    def _handshake(self, packet):
        if not packet.flags & UdpProtocol.FLAG_SYN:
            return

        if not packet.session == self.session_id:
            return

        response_code = packet.data[0]

        self.log.debug('Got response 0x{:02X} to handshake'.format(response_code))

        if response_code == 0x02:
            self.state = UdpProtocol.STATE_ESTABLISHED

        # Got a valid 2nd handshake packet, send back the third one
        response = Packet()
        response.flags = UdpProtocol.FLAG_ACK
        response.label = "_handshake"
        self._send_packet(response)

    def connect(self):
        if self.state != UdpProtocol.STATE_CLOSED:
            raise RuntimeError("Trying to open an connection that's already open")

        if not self.thread.is_alive():
            self.thread.start()

        # Reset internal state
        self.local_sequence_number = -1
        self.local_ack_number = 0
        self.remote_sequence_number = 0
        self.remote_ack_numbe = 0
        self.session_id = 0x1337
        self.enable_ack = False

        # Create first syn packet
        syn = Packet()
        syn.flags = UdpProtocol.FLAG_SYN
        syn.data = [
            0x01, 0x00,
            0x00, 0x00,
            0x00, 0x00,
            0x00, 0x00,
        ]
        self._send_packet(syn)
        self.state = UdpProtocol.STATE_SYN_SENT

    def receive_packet(self):
        while True:
            packet = self._receive_packet()

            if packet is True:
                continue
            if packet is None and not self.had_traffic:
                continue
            if packet is None and self.state == UdpProtocol.STATE_SYN_SENT:
                # No response in connect, retry connection
                self.state = UdpProtocol.STATE_CLOSED
                self.had_traffic = False
                self.connect()
                return None

            if packet is None:
                # When None is in the receive queue the socket has disconnected
                return None

            if self.mark_next_connected:
                self.mark_next_connected = False
                return ConnectionReady()

            if self.enable_ack and self.queue_trigger():
                return TransferQueueFlushed()

            if self.state == UdpProtocol.STATE_SYN_SENT:
                # Got response for the first handshake packet
                self.had_traffic = True
                self._handshake(packet)
            elif self.state == UdpProtocol.STATE_ESTABLISHED:
                if packet.length == 12:
                    # This is a control packet, deal with it in the transport layer
                    if not self.enable_ack:
                        # This is the first ACK from the mixer, after this we should send ACKs bac
                        self.enable_ack = True
                        # self.local_sequence_number = 0
                        ack = Packet()
                        ack.flags = UdpProtocol.FLAG_ACK
                        ack.acknowledgement_number = self.remote_sequence_number
                        ack.remote_sequence_number = 0x61
                        ack.label = 'initial ack after connection'
                        self._send_packet(ack)
                    # TODO: Implement other control packets, like request for retransmission

                    # Send queued up bulk traffic after the ack
                    if self.queue_trigger():
                        return TransferQueueFlushed()
                else:
                    # Data packet for the upper layer
                    return packet

    def send_packet(self, packet):
        self._send_packet(packet)


class UsbProtocol(BaseProtocol):
    STATE_INIT = 0
    STATE_CONNECTED = 1

    PRODUCTS = {
        0xbe49: "Atem Mini",
        0xbe55: "Atem Mini Pro",
        0xbe56: "Atem Mini Pro ISO",
        0xbe7c: "Atem Mini Extreme",
    }

    def __init__(self, port=None):
        super().__init__()
        port = port or "auto"
        self.port = port
        self.queue = Queue()

        self.log = logging.getLogger('USBTransport')
        self.thread = threading.Thread(None, self._rx_thread, "atem-usb", daemon=True)
        self.rx_queue = Queue(maxsize=1024)

        self.handle = UsbProtocol.find_device()
        self._detach_kernel()
        self.handle.set_configuration()

        self.rev = None
        self.init_done = False
        self.state = UsbProtocol.STATE_INIT
        self.buffer_use = 0

    @classmethod
    def device_exists(cls):
        return cls.find_device() is not None

    @classmethod
    def find_device(cls):
        for prod in UsbProtocol.PRODUCTS.keys():
            device = usb.core.find(idVendor=0x1edb, idProduct=prod)
            if device is not None:
                return device
        return None

    def _rx_thread(self):
        timer = 0
        while True:
            # Send buffer pressure info if more than a kilobyte of data is buffered
            diff = (time.time_ns() - timer) / (10 ** 9)
            if self.buffer_use > 1024 or diff < 0.01:
                buffer_blocks = min(1, self.buffer_use // 1024)
                if self.rev == 8.6:
                    self.write(0x03, struct.pack('>6xH', buffer_blocks))

            try:
                data = self.handle.read(0x82, 8192 * 4, timeout=0)
                timer = time.time_ns()
            except usb.USBError as e:
                if e.errno == 32:
                    self.log.fatal(f"USB failure on reading new packet: {e}")

                if e.errno == 110:
                    # Timeout happened, start reading again
                    continue
                break
            self.buffer_use += len(data)
            self.rx_queue.put(data)
        self.rx_queue.put(None)

    def _detach_kernel(self):
        try:
            for config in self.handle:
                for i in range(config.bNumInterfaces):
                    if self.handle.is_kernel_driver_active(i):
                        try:
                            self.handle.detach_kernel_driver(i)
                            self.log.debug('kernel driver detached')
                        except usb.core.USBError as e:
                            self.log.error('Could not detach kernel driver: ' + str(e))
        except usb.core.USBError as e:
            if e.errno == 13:
                raise PermissionError(e)

    def _send_packet(self, packet):
        raw = packet.to_usb()
        if self.rev == 8.6:
            raw = raw[4:]
            self.write(0x04, struct.pack('>6xH', len(raw)))
        self.handle.write(0x02, raw)

    def write(self, ptype, data, pid=1):
        raw = struct.pack('BBBB', 0, ptype, 0, 1) + data
        self.handle.write(0x02, raw)

    def read(self):
        result = self.handle.read(0x82, 16 * 1024)
        payload = bytes(result)
        header = struct.unpack_from('BBBB', payload, 0)
        return header[0], payload[4:]

    def connect(self):
        try:
            # Try the initial control transfer for firmware 8.5 and older
            self.handle.ctrl_transfer(0x21, 0, 0x0000, 2, [])
            # Start RX queue
            self.thread.start()
            self.rev = 8.5
            self.log.info("Using USB protocol version 8.5")
        except usb.core.USBError:
            # 8.6+ firmware raises a pipe error on the previous URB, try the newer handshake
            self.log.info('Using USB protocol version 8.6+')
            self.handle.reset()
            self.handle.clear_halt(0x02)
            self.handle.clear_halt(0x82)
            self.state = UsbProtocol.STATE_INIT

            # Start RX queue
            self.thread.start()

            # Init packet
            self.log.debug('Sending handshake packet 0x01')
            self.write(0x01, b'\x00\x01\x10\x92')
            self.rev = 8.6

    def receive_packet(self):
        if self.mark_next_connected:
            self.mark_next_connected = False
            return ConnectionReady()

        timeout = 0.5
        if self.rev < 8.6:
            timeout = 0.2

        while True:
            try:
                packet = self.rx_queue.get(timeout=timeout)
                time.sleep(0.01)
            except Empty:
                self.keep_alive()
                time.sleep(0.01)
                continue
            if packet is None:
                # Cannot read from USB anymore, device has disconnected
                return None
            self.buffer_use -= len(packet)

            if self.state == UsbProtocol.STATE_INIT:
                self.state = UsbProtocol.STATE_CONNECTED
                if self.rev == 8.6:
                    header = struct.unpack_from('BBBB', packet, 0)
                    self.log.debug(f'RX packet 0x{header[0]:x}')
                    self.log.debug(f'Send handshake packet 0x03')
                    self.write(0x03, b'x00\x00\x00\x00\x00\x00\x10\x00\x00\x00')
                    continue

            if self.rev == 8.6:
                header = struct.unpack_from('BBBB', packet, 0)
                packet_type = header[0]
                if packet_type == 0x03:
                    payload = struct.unpack('>BBBBH6x', packet)
                    self.log.warning(f"RX Buffer change received: {payload[0]}")
                elif packet_type == 0x04:
                    # Packet is ATEM protocol data.
                    payload_length, = struct.unpack_from('<H', packet, 4)

                    # Receive the actual ATEM data in the next packet
                    packet = b''
                    while len(packet) < payload_length:
                        chunk = self.rx_queue.get()
                        packet += chunk
                        self.buffer_use -= len(chunk)

                    if not self.mark_next_connected and not self.init_done:
                        self.mark_next_connected = True
                        self.init_done = True
                        self.log.debug("Mark connection ready after this")
                else:
                    self.log.warning(f"Unknown packet type received: {packet_type:#X}")
                    continue

            raw = bytes(packet)
            if self.rev == 8.6:
                packet = Packet()
                packet.data = raw
                return packet

            if len(raw) == 0:
                continue

            chunks = []
            while True:
                length, = Packet.STRUCT_USB.unpack_from(raw)
                chunks.append(raw[4:length + 4])
                raw = raw[length + 4:]
                if len(raw) == 0:
                    break

            packet = Packet()
            packet.data = b''.join(chunks)
            return packet

    def send_packet(self, packet):
        self._send_packet(packet)
        time.sleep(0.004)

    def keep_alive(self):
        try:
            self.handle.write(0x02, b'')
        except:
            pass


class TcpProtocol(BaseProtocol):
    STATE_INIT = 0
    STATE_AUTH = 1
    STATE_CONNECTED = 2

    STRUCT_HEADER = struct.Struct('!H')
    STRUCT_FIELD = struct.Struct('!H2x 4s')

    def __init__(self, url=None, host=None, port=None, username=None, password=None, device=None):
        super().__init__()
        if url is not None:
            part = urlparse(url)
            host = part.hostname
            port = part.port or 4532
            username = part.username
            password = part.password
            device = part.path[1:]
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.device = device

        self.sock = None
        self.state = TcpProtocol.STATE_INIT

        self.log = logging.getLogger('TcpTransport')

    def _send_packet(self, data):
        header = self.STRUCT_HEADER.pack(len(data))
        self.sock.sendall(header + data)

    def _receive_packet(self):
        try:
            header = self.sock.recv(2)
            datalength, = self.STRUCT_HEADER.unpack(header)
            data_left = datalength
            data = b''
            while data_left > 0:
                block = self.sock.recv(data_left)
                if len(block) == 0:
                    self.log.error("Connection closed")
                    return
                data_left -= len(block)
                data += block
        except:
            return None

        packet = Packet()
        packet.data = data
        return packet

    def decode_packet(self, data):
        offset = 0
        if len(data) < 8:
            raise ValueError("Packet too short")
        while offset < len(data):
            datalen, cmd = self.STRUCT_FIELD.unpack_from(data, offset)
            raw = data[offset + 8:offset + datalen]
            yield (cmd, raw)
            offset += datalen

    def list_to_packets(self, data):
        result = b''
        for key, value in data:
            result += self.STRUCT_FIELD.pack(len(value) + 8, key)
            result += value
        return result

    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((self.host, self.port))

        # Send magic packet to init the connection
        self._send_packet(self.list_to_packets([(b'*SW*', b'')]))

    def send_auth(self):
        if self.username is None or self.password is None:
            raise ValueError("Proxy requests AUTH but username or password is not set")
        self._send_packet(self.list_to_packets([
            (b'*USR', self.username.encode()),
            (b'*PWD', self.password.encode()),
        ]))

    def connect_device(self):
        self._send_packet(self.list_to_packets([
            (b'*DEV', self.device.encode()),
        ]))

    def receive_packet(self):
        while True:
            packet = self._receive_packet()
            if packet is None:
                continue
            if self.state == TcpProtocol.STATE_INIT:
                fields = list(self.decode_packet(packet.data))
                if fields[0][0] == b'AUTH':
                    self.send_auth()
                    continue
                elif fields[0][0] == b'*HW*':
                    self.connect_device()
                    self.state = TcpProtocol.STATE_CONNECTED
            elif self.state == TcpProtocol.STATE_CONNECTED:
                if packet is not None:
                    return packet

    def send_packet(self, packet):
        self._send_packet(packet.data)

    def upload(self, task):
        if not isinstance(task, TransferTask):
            raise ValueError()
        for packet in task.to_tcp():
            self._send_packet(self.list_to_packets([packet]))

    def download(self, task):
        pass



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import logging
import os
import struct
import urllib.request
from pathlib import Path

import gi
import pyatem

from gtk_switcher.colorwheel import ColorWheelWidget
from gtk_switcher.decorators import field
from pyatem.command import MediaplayerSelectCommand
from pyatem.field import InputPropertiesField
from pyatem.media import atem_to_rgb

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk, GdkPixbuf

gi.require_version('Handy', '1')
from gi.repository import Handy


class MediaPage:
    def __init__(self, builder):
        self.media_flow = builder.get_object('media_flow')
        self.model_media = builder.get_object('model_media')
        self.media_players = builder.get_object('media_players')
        self.media_slot = {}
        self.media_slot_name = {}
        self.media_slot_box = {}
        self.media_slot_progress = {}
        self.media_pixbuf = {}
        self.media_queue = []
        self.media_player_img = {}
        self.media_player_desc = {}
        self.media_last_upload = None

        self.media_context = None

        self.log_mp = logging.getLogger('MediaPage')

    @field('mediaplayer-slots')
    def on_mediaplayer_slots_change(self, data):
        for child in self.media_flow:
            child.destroy()
        self.model_changing = True
        self.model_media.clear()

        for i in range(0, data.stills):
            self.model_media.append([f"{i}", f"{i + 1}: "])

            slot = Gtk.Frame()
            slot.index = i
            slot.drag_dest_set(Gtk.DestDefaults.MOTION |
                               Gtk.DestDefaults.HIGHLIGHT | Gtk.DestDefaults.DROP,
                               [Gtk.TargetEntry.new("text/uri-list", 0, 80)], Gdk.DragAction.COPY)
            slot.connect('drag_data_received', self.on_media_file_dropped)
            slot.get_style_context().add_class('view')
            grid = Gtk.Grid()
            grid.set_column_spacing(8)
            grid.set_row_spacing(4)
            grid.set_margin_top(6)
            grid.set_margin_bottom(6)
            grid.set_margin_start(6)
            grid.set_margin_end(6)
            slot.add(grid)
            slot.get_style_context().add_class('media-slot')

            slot_number = Gtk.Label(label=str(i + 1))
            slot_number.get_style_context().add_class('dim-label')
            grid.attach(slot_number, 0, 0, 1, 1)

            slot_label = Gtk.Label("")
            slot_label.set_hexpand(True)
            slot_label.set_xalign(0.0)
            grid.attach(slot_label, 1, 0, 1, 1)

            slot_img = Gtk.Box()
            slot_img.get_style_context().add_class('mp-slot')
            slot_img.set_size_request(160, 120)
            grid.attach(slot_img, 0, 1, 2, 1)

            progress = Gtk.ProgressBar()
            progress.set_no_show_all(True)
            grid.attach(progress, 0, 2, 2, 1)

            self.media_slot[i] = slot
            self.media_slot_name[i] = slot_label
            self.media_slot_box[i] = slot_img
            self.media_slot_progress[i] = progress

            self.media_flow.add(slot)
        self.model_changing = False
        self.media_flow.show_all()

    def media_create_mediaplayers(self, count):
        for child in self.media_players:
            self.media_players.remove(child)

        width, height = self.connection.mixer.mixerstate['video-mode'].get_resolution()
        aw = 184
        ah = int(aw * height / width)

        for i in range(0, count):
            label = Gtk.Label(_("Media Player {}").format(i + 1))
            label.get_style_context().add_class('heading')
            self.media_players.add(label)
            frame = Gtk.Frame()
            frame.index = i
            frame.set_size_request(aw + 12, ah + 12)
            frame.get_style_context().add_class('view')
            self.media_players.add(frame)

            loaded = Gtk.Label("")
            loaded.set_margin_bottom(16)
            self.media_players.add(loaded)
            self.media_player_desc[i] = loaded

            frame.drag_dest_set(Gtk.DestDefaults.MOTION |
                                Gtk.DestDefaults.HIGHLIGHT | Gtk.DestDefaults.DROP,
                                [Gtk.TargetEntry.new("text/uri-list", 0, 80),
                                 Gtk.TargetEntry.new("text/x-slot-index", 0, 42)], Gdk.DragAction.COPY)
            frame.connect('drag_data_received', self.on_media_player_drag_data_received)

            img = Gtk.Image()
            img.set_margin_top(6)
            img.set_margin_bottom(6)
            img.set_margin_start(6)
            img.set_margin_end(6)
            self.media_player_img[i] = img
            frame.add(img)

        self.media_players.show_all()

    @field('mediaplayer-selected')
    def on_mediaplayer_media_source_change(self, data):
        if data.slot not in self.media_pixbuf:
            return

        if data.index not in self.media_player_img:
            return

        pixbuf = self.media_pixbuf[data.slot]
        width, height = self.connection.mixer.mixerstate['video-mode'].get_resolution()
        aw = 184
        thumb = pixbuf.scale_simple(aw, int(aw * height / width), GdkPixbuf.InterpType.BILINEAR)

        self.media_player_img[data.index].set_from_pixbuf(thumb)
        name = self.media_slot_name[data.slot].get_text()
        self.media_player_desc[data.index].set_text(f'Still {data.slot + 1}: {name}')

    @field('mediaplayer-file-info')
    def on_mediaplayer_file_info_change(self, data):
        if data.index not in self.media_slot_name:
            return
        self.set_class(self.media_slot[data.index], 'used', data.is_used)
        self.set_class(self.media_slot[data.index], 'empty', data.is_used)

        self.model_changing = True
        self.model_media[data.index][1] = f'{data.index + 1}: {data.name.decode()}'
        self.model_changing = False

        if self.media_last_upload is not None and data.index == self.media_last_upload:
            # Don't download the frame that was just uploaded
            self.media_last_upload = None
            return
        if data.is_used:
            self.media_slot_name[data.index].set_label(data.name.decode())
            if self.mainstack.get_visible_child_name() == 'media':
                self.connection.mixer.download(0, data.index)
            else:
                self.media_queue.append(data.index)
        else:
            if data.index in self.media_queue:
                self.media_queue.remove(data.index)
            self.media_slot_name[data.index].set_label("")
            for child in self.media_slot_box[data.index]:
                self.media_slot_box[data.index].remove(child)

    def on_media_transfer_progress(self, index, progress):
        if index not in self.media_slot_progress:
            return

        self.media_slot_progress[index].show()
        self.media_slot_progress[index].set_fraction(progress)

    def on_media_save(self, widget, target):
        frame = target.frame

        dialog = Gtk.FileChooserDialog(title="Saving frame", action=Gtk.FileChooserAction.SAVE)
        dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_SAVE,
            Gtk.ResponseType.OK,
        )

        for fmt in GdkPixbuf.Pixbuf.get_formats():
            if fmt.is_writable():
                f = Gtk.FileFilter()
                f.set_name(fmt.description)
                f.fname = fmt.name
                for ext in fmt.get_extensions():
                    f.add_pattern(f"*.{ext}")
                dialog.add_filter(f)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filter = dialog.get_filter()
            frame.savev(dialog.get_filename(), filter.fname)
        dialog.destroy()

    def on_media_context_menu(self, widget, event):
        self.log_mp.debug("Button", event.button)
        if event.button != 3:
            return
        self.media_context = Gtk.Menu.new()

        item = Gtk.MenuItem.new_with_label("Save image")
        item.connect('activate', self.on_media_save, widget)
        self.media_context.append(item)
        self.media_context.show_all()
        self.media_context.popup_at_pointer()

    def on_media_download_done(self, index, data):
        if index not in self.media_slot:
            return

        if len(data) == 0:
            return

        width, height = self.connection.mixer.mixerstate['video-mode'].get_resolution()
        raw = atem_to_rgb(data, width, height)

        # Pad the frame to the right size instead of failing hard when the transfer is corrupted
        if len(raw) != (width * 4 * height):
            missing = (width * 4 * height) - len(raw)
            raw += b'\0' * missing

        gdk_raw = GLib.Bytes.new(raw)
        pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(gdk_raw, GdkPixbuf.Colorspace.RGB, True, 8, width, height,
                                                 width * 4)
        self.media_pixbuf[index] = pixbuf
        aw = 184
        thumb = pixbuf.scale_simple(aw, int(aw * height / width), GdkPixbuf.InterpType.BILINEAR)
        thumb_img = Gtk.Image.new_from_pixbuf(thumb)
        eventbox = Gtk.EventBox()
        eventbox.add(thumb_img)
        eventbox.frame = pixbuf
        eventbox.index = index

        eventbox.drag_source_set(Gdk.ModifierType.BUTTON1_MASK, None, Gdk.DragAction.COPY | Gdk.DragAction.MOVE)
        eventbox.drag_source_add_uri_targets()
        eventbox.connect('drag_data_get', self.on_media_slot_drag_data_get)
        eventbox.connect('button-press-event', self.on_media_context_menu)

        if 'mediaplayer-selected' in self.connection.mixer.mixerstate:
            for mpi in self.connection.mixer.mixerstate['mediaplayer-selected']:
                mps = self.connection.mixer.mixerstate['mediaplayer-selected'][mpi]
                if mps.slot == index and mps.source_type == 1:
                    self.media_player_img[mpi].set_from_pixbuf(thumb)
                    name = self.media_slot_name[index].get_text()
                    self.media_player_desc[mpi].set_text(f'Still {index + 1}: {name}')

        self.media_slot_box[index].add(eventbox)
        self.media_slot_box[index].show_all()
        self.media_slot_progress[index].hide()

    def on_media_slot_drag_data_get(self, widget, drag_context, selection_data, info, time):
        atom = Gdk.Atom.intern('text/x-slot-index', False)
        data = struct.pack('B', widget.index)
        selection_data.set(atom, 42, data)

    def on_page_media_open(self):
        for index in self.media_queue:
            self.media_slot_progress[index].show()
            self.connection.mixer.download(0, index)
        self.media_queue = []

    def dnd_uri_to_path(self, uri):
        # Strip the various schemes from dropped uris
        path = ""
        if uri.startswith('file:\\\\\\'):
            path = uri[8:]
        elif uri.startswith('file://'):
            path = uri[7:]
        elif uri.startswith('file:'):
            path = uri[5:]

        path = urllib.request.url2pathname(path)
        path = path.strip('\r\n\x00')
        return path

    def on_media_player_drag_data_received(self, widget, context, x, y, selection, target_type, timestamp):
        index = widget.index
        # TODO: Accept files directly on the media player and upload to a free slot and switch
        if str(selection.get_data_type()) == 'text/x-slot-index':
            data = selection.get_data()
            slot, = struct.unpack('B', data)
            cmd = MediaplayerSelectCommand(index, still=slot)
            self.connection.mixer.send_commands([cmd])

    def on_media_file_dropped(self, widget, context, x, y, selection, target_type, timestamp):
        index = widget.index

        # URI List
        if target_type == 80:
            data = selection.get_data().decode().strip('\r\n\0')
            for uri in data.split('\r\n'):
                path = self.dnd_uri_to_path(uri)
                if not os.path.isfile(path):
                    self.log_mp.error(f"File does not exist: {path}")
                    continue
                self.media_slot_upload_file(index, path)
                index += 1

    def on_media_upload_done(self, store, slot):
        self.media_slot_progress[slot].hide()

        width, height = self.connection.mixer.mixerstate['video-mode'].get_resolution()
        pixbuf = self.media_pixbuf[slot]
        aw = 184
        thumb = pixbuf.scale_simple(aw, int(aw * height / width), GdkPixbuf.InterpType.BILINEAR)
        thumb_img = Gtk.Image.new_from_pixbuf(thumb)
        eventbox = Gtk.EventBox()
        eventbox.add(thumb_img)
        eventbox.frame = pixbuf
        eventbox.connect('button-press-event', self.on_media_context_menu)

        self.media_slot_box[slot].add(eventbox)
        self.media_slot_box[slot].show_all()

    def on_media_upload_progress(self, store, slot, percent, done, size):
        self.media_slot_progress[slot].set_fraction(percent / 100)

    def media_slot_upload_file(self, index, path):
        mode = self.connection.mixer.mixerstate['video-mode']
        width, height = mode.get_resolution()
        fname, ext = os.path.splitext(path)
        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(path, width, height, preserve_aspect_ratio=True)

        name = Path(path).stem
        premultiply = False
        if ext.lower() == '.png':
            premultiply = True
            
        self.media_slot_upload_pixbuf(index, pixbuf, name=name, premultiply=premultiply)

    def media_slot_upload_pixbuf(self, index, pixbuf, name=None, premultiply=False):
        self.log_mp.info(f"Uploading pixbuf to media slot {index}")
        mode = self.connection.mixer.mixerstate['video-mode']
        width, height = mode.get_resolution()
        aspect = width / height
        swidth = pixbuf.get_width()
        sheight = pixbuf.get_height()
        saspect = swidth / sheight

        if saspect < aspect:
            scale = sheight / height
            dest_y = 0
            dest_x = (width - (swidth * scale)) / 2
        else:
            scale = swidth / width
            dest_x = 0
            dest_y = (height - (sheight * scale)) / 2

        dest_w = swidth * scale
        dest_h = sheight * scale

        dest = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, width, height)
        pixbuf.scale(dest, dest_x, dest_y, dest_w, dest_h, dest_x, dest_y, scale, scale, GdkPixbuf.InterpType.BILINEAR)

        self.media_pixbuf[index] = pixbuf

        pixels = dest.get_pixels()
        frame = pyatem.media.rgb_to_atem(pixels, width, height, premultiply)
        self.media_slot_progress[index].show()
        self.media_slot[index].get_style_context().add_class('uploading')
        self.connection.mixer.upload(0, index, frame, name=name)
        self.media_last_upload = index



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import ctypes
import logging
import sys
import threading
import traceback

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import GLib

from pyatem.videohub import VideoHub


class VideoHubConnection(threading.Thread):
    def __init__(self, ip, connected, disconnected, input_change, output_change, route_change):
        threading.Thread.__init__(self)

        self.log = logging.getLogger('VideoHubConnection')

        self.id = None

        self._connected = [connected]
        self._disconnected = [disconnected]
        self._input_change = [input_change]
        self._output_change = [output_change]
        self._route_change = [route_change]

        self.ip = ip

        self.hub = None

        self.stop = False
        self.connected = False

        self.inputs = {}
        self.outputs = {}

    def run(self):
        self.hub = VideoHub(self.ip)
        self.hub.on('connect', self.do_connected)
        self.hub.on('disconnect', self.do_disconnected)
        self.hub.on('input-label-change', self.do_input_label_change)
        self.hub.on('input-status-change', self.do_input_status_change)
        self.hub.on('output-label-change', self.do_output_label_change)
        self.hub.on('route-change', self.do_route_change)

        try:
            self.hub.connect()
        except ConnectionError as e:
            self.log.error(f"Could not connect to VideoHub at {self.ip}: {e}")
            return
        except OSError as e:
            self.log.error(f"Could not connect to VideoHub at {self.ip}: {e}")
            return
        while not self.stop:
            try:
                self.hub.loop()
            except Exception as e:
                traceback.print_exc()
                self.log.error(repr(e))

    def do_disconnected(self, hub):
        self.connected = False
        for handler in self._disconnected:
            GLib.idle_add(handler, self.id)

    def do_connected(self, hub):
        for handler in self._connected:
            GLib.idle_add(handler, self.id)

    def do_input_label_change(self, hub, index, label):
        if index not in self.inputs:
            self.inputs[index] = {
                'status': None,
                'label': None
            }
        self.inputs[index]['label'] = label

        for handler in self._input_change:
            GLib.idle_add(handler, self.id, index, self.inputs)

    def do_input_status_change(self, hub, index, status):
        if index not in self.inputs:
            self.inputs[index] = {
                'status': None,
                'label': None
            }
        self.inputs[index]['status'] = status
        for handler in self._input_change:
            GLib.idle_add(handler, self.id, index, self.inputs)

    def do_output_label_change(self, hub, index, label):
        if index not in self.outputs:
            self.outputs[index] = {
                'source': None,
                'label': None
            }
        self.outputs[index]['label'] = label
        for handler in self._output_change:
            GLib.idle_add(handler, self.id, index, self.outputs)

    def do_route_change(self, hub, index, source):
        if index not in self.outputs:
            self.outputs[index] = {
                'source': None,
                'label': None
            }
        self.outputs[index]['source'] = source
        for handler in self._route_change:
            GLib.idle_add(handler, self.id, index, source)

    def change_route(self, index, source):
        self.hub.set_source(index, source)

    def get_id(self):
        # returns id of the respective thread
        if hasattr(self, '_thread_id'):
            return self._thread_id
        for id, thread in threading._active.items():
            if thread is self:
                return id

    def die(self):
        thread_id = self.get_id()
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id,
                                                         ctypes.py_object(SystemExit))
        if res > 1:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
            self.log.error('Exception raise failure')



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import shlex
import struct


class BaseMacroCommand:
    NAME = "Unknown action"
    TAG = "unknown-action"

    def __init__(self):
        self.actions = {}
        self.fields = {}
        self.lengths = {}
        self.widgets = {}

    def define_field(self, action, name, offset, packing):
        if action not in self.fields:
            self.fields[action] = []

        self.fields[action].append((offset, name, packing))

    def field_length(self, action, length):
        # Store action length without header
        self.lengths[action] = length - 4

    def add_action(self, action_id, raw):
        self.actions[action_id] = raw[4:]

    def add_widget(self, action_id, name, label, datatype, **kwargs):
        if action_id not in self.widgets:
            self.widgets[action_id] = []
        self.widgets[action_id].append((name, datatype, label, kwargs))

    def definition(self):
        pass

    def make_format(self, action):
        f = '<'
        current_offset = 0
        for offset, name, packing in sorted(self.fields[action]):
            if offset > current_offset:
                f += f'{offset - current_offset}x '
                current_offset = offset
            f += f'{packing} '
            current_offset += struct.calcsize(packing)
        if action in self.lengths:
            if current_offset < self.lengths[action]:
                f += '{}x'.format(self.lengths[action] - current_offset)
        return f

    def decode(self):
        self.definition()
        for action in self.fields:
            f = self.make_format(action)
            result = struct.unpack_from(f, self.actions[action], 0)
            for idx, field in enumerate(sorted(self.fields[action])):
                setattr(self, field[1], result[idx])

    def encode(self):
        result = b''
        for action in self.fields:
            f = self.make_format(action)
            fields = []
            for idx, field in enumerate(sorted(self.fields[action])):
                fields.append(getattr(self, field[1]))
            raw = struct.pack(f, *fields)
            header = struct.pack('<HH', self.lengths[action] + 4, action)
            result += header + raw
        return result

    def encode_script(self):
        result = self.__class__.TAG
        data = {}
        for action in self.fields:
            for offset, name, packing in self.fields[action]:
                data[name] = getattr(self, name)
        if len(data):
            parts = []
            for key in data:
                part = key + '='
                if isinstance(data[key], int) or isinstance(data[key], float) or isinstance(data[key], bool):
                    part += str(data[key])
                elif data[key] is None:
                    continue
                else:
                    part += '"' + str(data[key]).replace("\n", "\\n") + '"'
                parts.append(part)
            result += ' ' + ' '.join(parts)
        return result + "\n"

    def decode_script(self, raw_data):
        parts = shlex.split(raw_data, posix=False)
        for part in parts:
            key, value = part.split('=', maxsplit=1)
            if value.startswith('"'):
                value = value[1:-1]
            elif value == "False" or value == "True":
                value = value == "True"
            elif '.' in value:
                value = float(value)
            elif value.isnumeric():
                value = int(value)
            setattr(self, key, value)

    def __repr__(self):
        for a in self.actions:
            return '<action-unknown id={}>'.format(a)


class SleepMacroCommand(BaseMacroCommand):
    NAME = "Sleep"
    TAG = "sleep"

    def __init__(self):
        super().__init__()
        self.frames = None

    def definition(self):
        self.field_length(0x0007, 8)
        self.define_field(0x0007, 'frames', 0, 'H')
        self.add_widget(0x0007, 'frames', 'Duration', 'framecount')

    def __repr__(self):
        return '<sleep frames={}>'.format(self.frames)


class PreviewInputMacroCommand(BaseMacroCommand):
    NAME = "Preview input"
    TAG = "preview-input"

    def __init__(self):
        super().__init__()
        self.index = None
        self.source = None

    def definition(self):
        self.field_length(0x0003, 8)
        self.define_field(0x0003, 'index', 0, 'B')
        self.define_field(0x0003, 'source', 2, 'H')
        self.add_widget(0x0003, 'index', 'M/E unit', 'number', offset=1, min=1, max=4)
        self.add_widget(0x0003, 'source', 'Input', 'source', dataset='available_me')

    def __repr__(self):
        return '<preview-input me={} source={}>'.format(self.index, self.source)


class ProgramInputMacroCommand(BaseMacroCommand):
    NAME = "Program input"
    TAG = "program-input"

    def __repr__(self):
        return '<program-input>'


class TransitionWipeMacroCommand(BaseMacroCommand):
    NAME = "Transition wipe settings"
    TAG = "transition-wipe-settings"

    def __repr__(self):
        return '<transition-wipe>'



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
from pyatem.macrocommand import *
import pyatem.macrocommand as macromodule

macro_command_map = {
    0x0002: ProgramInputMacroCommand,
    0x0003: PreviewInputMacroCommand,
    0x0007: SleepMacroCommand,
    0x0014: TransitionWipeMacroCommand,
    0x0015: TransitionWipeMacroCommand,
    0x0016: TransitionWipeMacroCommand
}


def decode_macro(raw):
    result = []
    offset = 0

    last_command = None
    while offset < len(raw):
        length, command_id = struct.unpack_from('<HH', raw, offset)
        command_raw = raw[offset:offset + length]
        if command_id in macro_command_map:
            command = macro_command_map[command_id]
            if command == last_command:
                result[-1].add_action(command_id, command_raw)
            else:
                decoder = command()
                decoder.add_action(command_id, command_raw)
                result.append(decoder)
        else:
            decoder = BaseMacroCommand()
            decoder.add_action(command_id, command_raw)
            result.append(decoder)
        offset += length

    for decoder in result:
        decoder.decode()

    return result


def encode_macro(actions):
    result = b''
    for action in actions:
        result += action.encode()
    return result


def encode_macroscript(actions):
    result = ''
    for action in actions:
        result += action.encode_script()
    return result


def decode_macroscript(script):
    classmap = {}
    for name, cls in macromodule.__dict__.items():
        if isinstance(cls, type) and name != "BaseMacroCommand":
            classmap[cls.TAG] = cls

    result = []
    for line in script.splitlines(keepends=False):
        name, data = line.split(' ', maxsplit=1)
        if name in classmap:
            inst = classmap[name]()
        else:
            inst = BaseMacroCommand()
        inst.decode_script(data)
        result.append(inst)
    return result


if __name__ == '__main__':

    with open('/workspace/usb-65535-0.bin', 'rb') as handle:
        raw_macro = handle.read()

    result = decode_macro(raw_macro)
    print(result)

    encoded = encode_macroscript(result)
    print(encoded)
    decoded = decode_macroscript(encoded)
    print(decoded)



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
from gi.repository import Gtk, GObject, Gdk

from gtk_switcher.template_i18n import TemplateLocale
from pyatem.command import KeyFillCommand, KeyPropertiesDveCommand, KeyTypeCommand, KeyCutCommand, \
    KeyPropertiesLumaCommand, KeyerKeyframeSetCommand, KeyerKeyframeRunCommand, KeyPropertiesAdvancedChromaCommand, \
    KeyPropertiesAdvancedChromaColorpickerCommand
from pyatem.field import TransitionSettingsField, KeyPropertiesDveField, KeyPropertiesLumaField, \
    KeyPropertiesAdvancedChromaField


@TemplateLocale(resource_path='/nl/brixit/switcher/ui/upstream-keyer.glade')
class UpstreamKeyer(Gtk.Frame):
    __gtype_name__ = 'UpstreamKeyer'

    keyer_stack = Gtk.Template.Child()
    mask_top = Gtk.Template.Child()
    mask_bottom = Gtk.Template.Child()
    mask_left = Gtk.Template.Child()
    mask_right = Gtk.Template.Child()
    mask_en = Gtk.Template.Child()

    luma_fill = Gtk.Template.Child()
    luma_key = Gtk.Template.Child()
    luma_premultiplied = Gtk.Template.Child()
    luma_clip = Gtk.Template.Child()
    luma_clip_adj = Gtk.Template.Child()
    luma_gain = Gtk.Template.Child()
    luma_gain_adj = Gtk.Template.Child()
    luma_invert = Gtk.Template.Child()

    chroma_sample = Gtk.Template.Child()
    chroma_sample_pick = Gtk.Template.Child()
    chroma_sample_preview = Gtk.Template.Child()

    chroma_fill = Gtk.Template.Child()
    chroma_foreground_adj = Gtk.Template.Child()
    chroma_background_adj = Gtk.Template.Child()
    chroma_edge_adj = Gtk.Template.Child()
    chroma_spill_adj = Gtk.Template.Child()
    chroma_flare_adj = Gtk.Template.Child()
    chroma_brightness_adj = Gtk.Template.Child()
    chroma_contrast_adj = Gtk.Template.Child()
    chroma_saturation_adj = Gtk.Template.Child()
    chroma_red_adj = Gtk.Template.Child()
    chroma_green_adj = Gtk.Template.Child()
    chroma_blue_adj = Gtk.Template.Child()

    dve_fill = Gtk.Template.Child()
    dve_shadow_en = Gtk.Template.Child()

    dve_pos_x_adj = Gtk.Template.Child()
    dve_pos_y_adj = Gtk.Template.Child()
    dve_size_x_adj = Gtk.Template.Child()
    dve_size_y_adj = Gtk.Template.Child()
    dve_light_angle_adj = Gtk.Template.Child()
    dve_light_altitude_adj = Gtk.Template.Child()

    dve_border_en = Gtk.Template.Child()
    dve_border_color = Gtk.Template.Child()
    dve_border_opacity = Gtk.Template.Child()
    dve_border_outer_width_adj = Gtk.Template.Child()
    dve_border_inner_width_adj = Gtk.Template.Child()
    dve_border_outer_soften_adj = Gtk.Template.Child()
    dve_border_inner_soften_adj = Gtk.Template.Child()
    dve_border_bevel_position_adj = Gtk.Template.Child()
    dve_border_bevel_soften_adj = Gtk.Template.Child()

    dve_set_a = Gtk.Template.Child()
    dve_set_b = Gtk.Template.Child()

    def __init__(self, index, keyer, connection):
        self.index = index
        self.keyer = keyer
        self.connection = connection

        self.model_changing = False
        self.slider_held = False

        super(Gtk.Frame, self).__init__()
        self.init_template()

    def __repr__(self):
        return '<UpstreamKeyer me={} keyer={}>'.format(self.index, self.keyer)

    def set_class(self, widget, classname, state):
        if state:
            widget.get_style_context().add_class(classname)
        else:
            widget.get_style_context().remove_class(classname)

    def set_fill_model(self, model):
        self.model_changing = True
        self.dve_fill.set_model(model)
        self.luma_fill.set_model(model)
        self.chroma_fill.set_model(model)
        self.model_changing = False

    def set_key_model(self, model):
        self.model_changing = True
        self.luma_key.set_model(model)
        self.model_changing = False

    def on_key_properties_base_change(self, data):
        self.model_changing = True
        self.dve_fill.set_active_id(str(data.fill_source))
        self.luma_fill.set_active_id(str(data.fill_source))
        self.chroma_fill.set_active_id(str(data.fill_source))
        self.luma_key.set_active_id(str(data.key_source))

        if data.type == 0:
            self.keyer_stack.set_visible_child_name('key_luma')
        elif data.type == 1:
            self.keyer_stack.set_visible_child_name('key_chroma')
        elif data.type == 2:
            self.keyer_stack.set_visible_child_name('key_pattern')
        elif data.type == 3:
            self.keyer_stack.set_visible_child_name('key_dve')

        self.model_changing = False

        if data.type != 3:
            self.set_class(self.mask_en, 'active', data.mask_enabled)

            self.mask_top.set_text(str(data.mask_top / 1000))
            self.mask_bottom.set_text(str(data.mask_bottom / 1000))
            self.mask_left.set_text(str(data.mask_left / 1000))
            self.mask_right.set_text(str(data.mask_right / 1000))

    def on_key_properties_luma_change(self, data):
        if not isinstance(data, KeyPropertiesLumaField):
            return

        self.set_class(self.luma_premultiplied, 'active', data.premultiplied)
        self.set_class(self.luma_invert, 'active', data.key_inverted)

        self.luma_clip.set_sensitive(not data.premultiplied)
        self.luma_gain.set_sensitive(not data.premultiplied)

        self.luma_clip_adj.set_value(data.clip)
        self.luma_gain_adj.set_value(data.gain)

    def on_key_properties_dve_change(self, data):
        if not isinstance(data, KeyPropertiesDveField):
            return

        self.set_class(self.dve_shadow_en, 'active', data.shadow_enabled)
        self.set_class(self.dve_border_en, 'active', data.border_enabled)

        self.dve_pos_x_adj.set_value(data.pos_x)
        self.dve_pos_y_adj.set_value(data.pos_y)
        self.dve_size_x_adj.set_value(data.size_x)
        self.dve_size_y_adj.set_value(data.size_y)
        self.dve_light_angle_adj.set_value(data.light_angle)
        self.dve_light_altitude_adj.set_value(data.light_altitude)

        self.dve_border_opacity.set_value(data.border_opacity)
        self.dve_border_outer_width_adj.set_value(data.border_outer_width)
        self.dve_border_inner_width_adj.set_value(data.border_inner_width)
        self.dve_border_outer_soften_adj.set_value(data.border_outer_softness)
        self.dve_border_inner_soften_adj.set_value(data.border_inner_softness)
        self.dve_border_bevel_position_adj.set_value(data.border_bevel_position)
        self.dve_border_bevel_soften_adj.set_value(data.border_bevel_softness)

        r, g, b = data.get_border_color_rgb()
        color = Gdk.RGBA()
        color.red = r
        color.green = g
        color.blue = b
        color.alpha = 1.0
        self.dve_border_color.set_rgba(color)

        self.set_class(self.mask_en, 'active', data.mask_enabled)

        self.mask_top.set_text(str(data.mask_top / 1000))
        self.mask_bottom.set_text(str(data.mask_bottom / 1000))
        self.mask_left.set_text(str(data.mask_left / 1000))
        self.mask_right.set_text(str(data.mask_right / 1000))

    def on_advanced_chroma_change(self, data):
        if not isinstance(data, KeyPropertiesAdvancedChromaField):
            return

        self.model_changing = True
        self.chroma_foreground_adj.set_value(data.foreground)
        self.chroma_background_adj.set_value(data.background)
        self.chroma_edge_adj.set_value(data.key_edge)
        self.chroma_spill_adj.set_value(data.spill_suppress)
        self.chroma_flare_adj.set_value(data.flare_suppress)
        self.chroma_brightness_adj.set_value(data.brightness)
        self.chroma_contrast_adj.set_value(data.contrast)
        self.chroma_saturation_adj.set_value(data.saturation)
        self.chroma_red_adj.set_value(data.red)
        self.chroma_green_adj.set_value(data.green)
        self.chroma_blue_adj.set_value(data.blue)
        self.model_changing = False

    def on_chroma_picker_change(self, data):
        self.set_class(self.chroma_sample_pick, 'active', data.cursor)
        self.set_class(self.chroma_sample_preview, 'active', data.preview)

        r, g, b = data.get_rgb()
        color = Gdk.RGBA()
        color.red = r
        color.green = g
        color.blue = b
        color.alpha = 1.0

        self.chroma_sample.set_rgba(color)

    @Gtk.Template.Callback()
    def on_fill_changed(self, widget, *args):
        if self.model_changing:
            return
        source = int(widget.get_active_id())
        cmd = KeyFillCommand(index=self.index, keyer=self.keyer, source=source)
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_key_changed(self, widget, *args):
        if self.model_changing:
            return
        source = int(widget.get_active_id())
        cmd = KeyCutCommand(index=self.index, keyer=self.keyer, source=source)
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_slider_press(self, *args):
        self.slider_held = True

    @Gtk.Template.Callback()
    def on_slider_release(self, *args):
        self.slider_held = False

    @Gtk.Template.Callback()
    def on_dve_pos_x_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, pos_x=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_pos_y_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, pos_y=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_size_x_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, size_x=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_size_y_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, size_y=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_shadow_en_clicked(self, widget):
        state = widget.get_style_context().has_class('active')
        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, shadow_enabled=not state)
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_light_angle_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, angle=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_light_altitude_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, altitude=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_border_en_clicked(self, widget):
        state = widget.get_style_context().has_class('active')
        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, border_enabled=not state)
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_border_color_color_set(self, widget):
        color = widget.get_rgba()
        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer)
        cmd.set_border_color_rgb(color.red, color.green, color.blue)
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_border_opacity_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, border_opacity=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_border_outer_width_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, outer_width=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_border_inner_width_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, inner_width=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_border_outer_soften_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, outer_softness=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_border_inner_soften_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, inner_softness=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_border_bevel_soften_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, bevel_softness=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_border_bevel_position_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, bevel_position=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_keyer_stack_visible_child_notify(self, widget, *args):
        if self.model_changing:
            return
        state = widget.get_visible_child_name()
        if state == 'key_luma':
            cmd = KeyTypeCommand(index=self.index, keyer=self.keyer, type=KeyTypeCommand.LUMA)
        elif state == 'key_chroma':
            cmd = KeyTypeCommand(index=self.index, keyer=self.keyer, type=KeyTypeCommand.CHROMA)
        elif state == 'key_pattern':
            cmd = KeyTypeCommand(index=self.index, keyer=self.keyer, type=KeyTypeCommand.PATTERN)
        elif state == 'key_dve':
            cmd = KeyTypeCommand(index=self.index, keyer=self.keyer, type=KeyTypeCommand.DVE)

        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_luma_premultiplied_clicked(self, widget, *args):
        state = widget.get_style_context().has_class('active')
        cmd = KeyPropertiesLumaCommand(index=self.index, keyer=self.keyer, premultiplied=not state)
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_luma_invert_clicked(self, widget, *args):
        state = widget.get_style_context().has_class('active')
        cmd = KeyPropertiesLumaCommand(index=self.index, keyer=self.keyer, invert_key=not state)
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_luma_clip_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesLumaCommand(index=self.index, keyer=self.keyer, clip=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_luma_gain_adj_value_changed(self, widget):
        if not self.slider_held:
            return

        cmd = KeyPropertiesLumaCommand(index=self.index, keyer=self.keyer, gain=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_lock_toggled(self, widget):
        self.set_class(widget, 'program', not widget.get_active)
        self.dve_set_a.set_sensitive(not widget.get_active())
        self.dve_set_b.set_sensitive(not widget.get_active())

    @Gtk.Template.Callback()
    def on_dve_set_a_clicked(self, widget):
        cmd = KeyerKeyframeSetCommand(index=self.index, keyer=self.keyer, keyframe='A')
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_set_b_clicked(self, widget):
        cmd = KeyerKeyframeSetCommand(index=self.index, keyer=self.keyer, keyframe='B')
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_run_full_clicked(self, widget):
        cmd = KeyerKeyframeRunCommand(index=self.index, keyer=self.keyer, run_to='Full')
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_run_a_clicked(self, widget):
        cmd = KeyerKeyframeRunCommand(index=self.index, keyer=self.keyer, run_to='A')
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_dve_run_b_clicked(self, widget):
        cmd = KeyerKeyframeRunCommand(index=self.index, keyer=self.keyer, run_to='B')
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_pick_clicked(self, widget):
        state = widget.get_style_context().has_class('active')
        cmd = KeyPropertiesAdvancedChromaColorpickerCommand(index=self.index, keyer=self.keyer, cursor=not state)
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_preview_clicked(self, widget):
        state = widget.get_style_context().has_class('active')
        cmd = KeyPropertiesAdvancedChromaColorpickerCommand(index=self.index, keyer=self.keyer, preview=not state)
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_foreground_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, foreground=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_background_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, background=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_edge_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, key_edge=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_spill_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, spill=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_flare_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, flare=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_brightness_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, brightness=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_contrast_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, contrast=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_saturation_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, saturation=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_red_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, red=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_green_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, green=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_chroma_blue_adj_value_changed(self, widget):
        if not self.slider_held:
            return
        cmd = KeyPropertiesAdvancedChromaCommand(index=self.index, keyer=self.keyer, blue=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    @Gtk.Template.Callback()
    def on_mask_enable_clicked(self, widget):
        state = widget.get_style_context().has_class('active')
        cmd = KeyPropertiesDveCommand(index=self.index, keyer=self.keyer, mask_enabled=not state)
        self.connection.mixer.send_commands([cmd])



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import json

import gi

from pyatem.command import MultiviewInputCommand, VideoModeCommand, AutoInputVideoModeCommand, SaveStartupStateCommand, \
    ClearStartupStateCommand
from pyatem.field import InputPropertiesField

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk

gi.require_version('Handy', '1')
from gi.repository import Handy


class PreferencesWindow:
    def __init__(self, parent, application, connection):
        self.application = application
        self.connection = connection
        self.settings = Gio.Settings.new('nl.brixit.Switcher')
        self.model_changing = False
        self.config = {}

        builder = Gtk.Builder()
        builder.set_translation_domain("openswitcher")
        builder.add_from_resource('/nl/brixit/switcher/ui/preferences.glade')
        builder.connect_signals(self)
        css = Gio.resources_lookup_data("/nl/brixit/switcher/ui/style.css", 0)

        self.provider = Gtk.CssProvider()
        self.provider.load_from_data(css.get_data())

        self.window = builder.get_object("window")
        self.window.set_application(self.application)

        self.multiview_window = []

        self.model_me1 = builder.get_object("model_me1")
        self.model_aux = builder.get_object("model_aux")
        self.model_route_inputs = Gtk.ListStore(str, str)

        # Load requested view
        self.mainstack = builder.get_object("mainstack")
        self.settingsstack = builder.get_object("settingsstack")

        self.video_mode = builder.get_object('video_mode')
        self.current_video_mode = builder.get_object('current_video_mode')
        self.model_video_mode = builder.get_object('model_video_mode')
        self.multiview_mode = builder.get_object('multiview_mode')
        self.current_multiview_mode = builder.get_object('current_multiview_mode')
        self.model_multiview_mode = builder.get_object('model_multiview_mode')
        self.downconvert_mode = builder.get_object('downconvert_mode')
        self.current_downconvert_mode = builder.get_object('current_downconvert_mode')
        self.model_downconvert_mode = builder.get_object('model_downconvert_mode')

        self.multiview_layout = builder.get_object("multiview_layout")
        self.multiview_tl = builder.get_object("multiview_tl")
        self.multiview_tr = builder.get_object("multiview_tr")
        self.multiview_bl = builder.get_object("multiview_bl")
        self.multiview_br = builder.get_object("multiview_br")
        self.multiview_swap = builder.get_object("multiview_swap")
        self.multiview_layout = builder.get_object("multiview_layout")

        self.videohubs = builder.get_object("videohubs")

        self.apply_css(self.window, self.provider)

        self.window.set_transient_for(parent)
        self.window.set_modal(True)
        self.load_models()
        self.load_preferences()
        self.load_config()
        self.connection.mixer.on('change:multiviewer-properties:*', self.make_multiviewer)
        self.connection.mixer.on('change:multiviewer-input:*', self.update_multiviewer_input)
        self.connection.mixer.on('change:video-mode', self.update_mode_models)
        self.window.show_all()

    def load_config(self):
        from gtk_switcher.videohub import VideoHub

        connections = self.settings.get_string('connections')
        connections = json.loads(connections)
        if self.connection.ip in connections:
            self.config = connections[self.connection.ip]
        else:
            self.config = {
                'videohubs': {}
            }

        for ip in self.config['videohubs']:
            config = self.config['videohubs'][ip]
            hub = VideoHub()
            hub.set_input_model(self.model_route_inputs)
            hub.connect('config-changed', self.on_videohub_config_changed)
            hub.connect('ip-changed', self.on_videohub_ip_changed)
            hub.connect('deleted', self.on_videohub_deleted)
            hub.load(config)
            self.videohubs.add(hub)
        self.videohubs.show_all()

    def on_videohub_config_changed(self, widget, config):
        config = json.loads(config)
        self.config['videohubs'][config['ip']] = config
        self.save_config()

    def on_videohub_ip_changed(self, widget, old_ip, new_ip):
        if old_ip is None:
            return

        hub_config = self.config['videohubs'][old_ip]
        del self.config['videohubs'][old_ip]
        self.config['videohubs'][new_ip] = hub_config
        self.save_config()

    def on_videohub_deleted(self, widget, ip):
        del self.config['videohubs'][ip]
        self.videohubs.remove(widget)
        self.save_config()

    def save_config(self):
        connections = self.settings.get_string('connections')
        connections = json.loads(connections)
        connections[self.connection.ip] = self.config
        connections = json.dumps(connections)
        self.settings.set_string('connections', connections)

    def load_models(self):
        inputs = self.connection.mixer.mixerstate['input-properties']
        self.model_changing = True
        self.model_route_inputs.clear()
        self.model_route_inputs.append(["", ""])
        for i in inputs.values():
            if i.available_me1:
                self.model_me1.append([str(i.index), i.name])

            if i.available_aux:
                self.model_aux.append([str(i.index), i.name])

            if i.port_type == InputPropertiesField.PORT_EXTERNAL:
                self.model_route_inputs.append([str(i.index), f"{i.index}: {i.name}"])

        self.model_changing = False

        self.update_mode_models(None, update_active=True)

    def update_mode_models(self, arg, update_active=False):
        current_mode = self.connection.mixer.mixerstate['video-mode']
        if 'auto-input-video-mode' in self.connection.mixer.mixerstate:
            am = self.connection.mixer.mixerstate['auto-input-video-mode']
            has_auto = True
            automode = am.enabled
        else:
            has_auto = False
            automode = False

        old_video_mode = self.video_mode.get_active_id()
        old_multiview_mode = self.multiview_mode.get_active_id()
        old_downscale_mode = self.downconvert_mode.get_active_id()

        self.current_video_mode.set_text(current_mode.get_label())
        self.model_video_mode.clear()
        self.model_multiview_mode.clear()
        self.model_downconvert_mode.clear()

        if has_auto:
            self.model_video_mode.append(['auto', 'Auto'])

        cmi = {'multiview': [], 'downscale': []}
        for mode in self.connection.mixer.mixerstate['video-mode-capability'].modes:
            self.model_video_mode.append([str(mode['modenum']), mode['mode'].get_label()])
            if mode['modenum'] == current_mode.mode:
                cmi = mode
                for mv_mode in mode['multiview']:
                    self.model_multiview_mode.append([str(mv_mode.mode), mv_mode.get_label()])
                for dc_mode in mode['downscale']:
                    self.model_downconvert_mode.append([str(dc_mode.mode), dc_mode.get_label()])

        if update_active:
            if automode:
                self.video_mode.set_active_id('auto')
            else:
                self.video_mode.set_active_id(str(current_mode.mode))

            if len(cmi['multiview']) == 1:
                self.multiview_mode.set_active_id(str(cmi['multiview'][0].mode))

            if len(cmi['downscale']) == 1:
                self.downconvert_mode.set_active_id(str(cmi['downscale'][0].mode))

        else:
            self.video_mode.set_active_id(old_video_mode)
            self.multiview_mode.set_active_id(old_multiview_mode)
            self.downconvert_mode.set_active_id(old_downscale_mode)

    def on_video_mode_selection_changed(self, widget):
        self.model_multiview_mode.clear()
        self.model_downconvert_mode.clear()
        current_mode = None
        for mode in self.connection.mixer.mixerstate['video-mode-capability'].modes:
            if str(mode['modenum']) == self.video_mode.get_active_id():
                current_mode = mode
                for mv_mode in mode['multiview']:
                    self.model_multiview_mode.append([str(mv_mode.mode), mv_mode.get_label()])
                for dc_mode in mode['downscale']:
                    self.model_downconvert_mode.append([str(dc_mode.mode), dc_mode.get_label()])

        if current_mode:
            if len(current_mode['multiview']) == 1:
                self.multiview_mode.set_active_id(str(current_mode['multiview'][0].mode))
            if len(current_mode['downscale']) == 1:
                self.downconvert_mode.set_active_id(str(current_mode['downscale'][0].mode))

    def load_preferences(self):
        state = self.connection.mixer.mixerstate

        if 'multiviewer-properties' in state and 'multiviewer-input' in state:
            self.make_multiviewer()

    def update_multiviewer_input(self, input):
        state = self.connection.mixer.mixerstate

        if 'multiviewer-properties' in state and 'multiviewer-input' in state:
            self.make_multiviewer()

    def make_multiviewer(self, *args):
        state = self.connection.mixer.mixerstate
        multiviewer = state['multiviewer-properties'][0]
        self.multiview_window = []
        for widget in self.multiview_layout:
            self.multiview_layout.remove(widget)

        sideways = multiviewer.layout == 5 or multiviewer.layout == 10

        if sideways:
            if not multiviewer.top_left_small:
                self.make_multiview_window(0, 0)
            if not multiviewer.bottom_left_small:
                self.make_multiview_window(0, 1)
            if not multiviewer.top_right_small:
                self.make_multiview_window(1, 0)
            if not multiviewer.bottom_right_small:
                self.make_multiview_window(1, 1)
        else:
            if not multiviewer.top_left_small:
                self.make_multiview_window(0, 0)
            if not multiviewer.top_right_small:
                self.make_multiview_window(1, 0)
            if not multiviewer.bottom_left_small:
                self.make_multiview_window(0, 1)
            if not multiviewer.bottom_right_small:
                self.make_multiview_window(1, 1)

        if multiviewer.top_left_small:
            self.make_split_multiview_window(0, 0, False)
        if multiviewer.top_right_small:
            self.make_split_multiview_window(1, 0, False)
        if multiviewer.top_left_small:
            self.make_split_multiview_window(0, 0, True)
        if multiviewer.top_right_small:
            self.make_split_multiview_window(1, 0, True)

        if multiviewer.bottom_left_small:
            self.make_split_multiview_window(0, 1, False)
        if multiviewer.bottom_right_small:
            self.make_split_multiview_window(1, 1, False)
        if multiviewer.bottom_left_small:
            self.make_split_multiview_window(0, 1, True)
        if multiviewer.bottom_right_small:
            self.make_split_multiview_window(1, 1, True)

        for index, window in enumerate(self.multiview_window):
            window.index = index

        self.multiview_tl.set_active(not multiviewer.top_left_small)
        self.multiview_tr.set_active(not multiviewer.top_right_small)
        self.multiview_bl.set_active(not multiviewer.bottom_left_small)
        self.multiview_br.set_active(not multiviewer.bottom_right_small)
        self.multiview_swap.set_active(multiviewer.flip)
        self.multiview_layout.show_all()

    def make_split_multiview_window(self, x, y, second=False):
        x *= 2
        y *= 2
        if not second:
            self.make_multiview_window(x, y, 1, 1)
            self.make_multiview_window(x + 1, y, 1, 1)
        else:
            self.make_multiview_window(x, y + 1, 1, 1)
            self.make_multiview_window(x + 1, y + 1, 1, 1)

    def make_multiview_window(self, x, y, w=2, h=2):
        x *= w
        y *= h
        routable = self.connection.mixer.mixerstate['topology'].multiviewer_routable
        frame = Gtk.Frame()
        frame.get_style_context().add_class('view')

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        frame.add(box)
        box.set_margin_top(8)
        box.set_margin_bottom(8)
        box.set_margin_left(8)
        box.set_margin_right(8)

        index = len(self.multiview_window)
        if index in self.connection.mixer.mixerstate['multiviewer-input'][0]:
            input = self.connection.mixer.mixerstate['multiviewer-input'][0][index]
            ip = self.connection.mixer.mixerstate['input-properties'][input.source]

            if routable:
                input_select = Gtk.ComboBox.new_with_model(self.model_aux)
                input_select.set_entry_text_column(1)
                input_select.set_id_column(0)
                input_select.window = index
                renderer = Gtk.CellRendererText()
                input_select.pack_start(renderer, True)
                input_select.add_attribute(renderer, "text", 1)
                input_select.set_active_id(str(input.source))
                input_select.set_margin_bottom(16)
                input_select.connect('changed', self.on_multiview_window_changed)
                box.add(input_select)
            else:
                input_label = Gtk.Label(ip.name)
                input_label.set_margin_bottom(16)
                box.add(input_label)

            buttonbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
            box.pack_end(buttonbox, False, False, 0)
            if input.vu:
                vu = self.connection.mixer.mixerstate['multiviewer-vu'][0][index]
                icon = Gtk.Image.new_from_resource("/nl/brixit/switcher/icons/multiview-vu.svg")
                vubutton = Gtk.ToggleButton(image=icon)
                vubutton.set_active(vu.enabled)
                buttonbox.add(vubutton)
            if input.safearea:
                sa = self.connection.mixer.mixerstate['multiviewer-safe-area'][0][index]
                icon = Gtk.Image.new_from_resource("/nl/brixit/switcher/icons/multiview-safearea.svg")
                sabutton = Gtk.ToggleButton(image=icon)
                sabutton.set_active(sa.enabled)
                buttonbox.add(sabutton)

        self.multiview_layout.attach(frame, x, y, w, h)
        self.multiview_window.append(frame)

    def on_multiview_window_changed(self, widget):
        if self.model_changing:
            return
        cmd = MultiviewInputCommand(index=0, window=widget.window, source=int(widget.get_active_id()))
        self.connection.mixer.send_commands([cmd])

    def apply_css(self, widget, provider):
        Gtk.StyleContext.add_provider(widget.get_style_context(),
                                      provider,
                                      Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        if isinstance(widget, Gtk.Container):
            widget.forall(self.apply_css, provider)

    def on_set_video_mode_clicked(self, widget):
        cmds = []
        if 'auto-input-video-mode' in self.connection.mixer.mixerstate:
            am = self.connection.mixer.mixerstate['auto-input-video-mode']
            if am.enabled and self.video_mode.get_active_id() != 'auto':
                cmds.append(AutoInputVideoModeCommand(False))
            if not am.enabled and self.video_mode.get_active_id() == 'auto':
                cmd = AutoInputVideoModeCommand(True)
                self.connection.mixer.send_commands([cmd])
                return

        cmds.append(VideoModeCommand(int(self.video_mode.get_active_id())))
        self.connection.mixer.send_commands(cmds)

    def on_add_videohub_clicked(self, widget, *args):
        from gtk_switcher.videohub import VideoHub

        hub = VideoHub()
        hub.set_input_model(self.model_route_inputs)
        hub.connect('config-changed', self.on_videohub_config_changed)
        hub.connect('ip-changed', self.on_videohub_ip_changed)
        hub.connect('deleted', self.on_videohub_deleted)
        self.videohubs.add(hub)
        self.videohubs.show_all()

    def on_save_startup_state(self, widget):
        self.connection.mixer.send_commands([SaveStartupStateCommand()])

    def on_clear_startup_state(self, widget):
        self.connection.mixer.send_commands([ClearStartupStateCommand()])


import argparse
import sys
import time

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib
import logging

from openswitcher_proxy.error import RecoverableError
from openswitcher_proxy.frontend_httpapi import HttpApiFrontendThread
from openswitcher_proxy.frontend_midi import MidiFrontendThread
from openswitcher_proxy.frontend_status import StatusFrontendThread
from openswitcher_proxy.frontend_tcp import TcpFrontendThread
from openswitcher_proxy.frontend_mqtt import MqttFrontendThread
from openswitcher_proxy.hardware import HardwareThread

logging.basicConfig(
    format="%(asctime)s [%(levelname)-8s %(threadName)-15s] %(message)s",
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S',
)

threads = []
nthreads = {}


def run(config_path):
    with open(config_path, 'rb') as handle:
        config = tomllib.load(handle)
    logging.info('Loading config file ' + config_path)
    if 'hardware' in config:
        nthreads['hardware'] = {}
        for hardware in config['hardware']:
            logging.info(f'  hardware: {hardware["id"]} ({hardware["label"]})')
            t = HardwareThread(hardware)
            t.daemon = True
            threads.append(t)
            nthreads['hardware'][hardware['id']] = t
            t.start()

    if 'frontend' in config:
        nthreads['frontend'] = {}
        for frontend in config['frontend']:
            if 'host' in frontend:
                connection = frontend['host']
            elif 'bind' in frontend:
                connection = frontend['bind']
            else:
                connection = 'n/a'
                logging.error(f'  Frontend is missing bind or host option')
            logging.info(f'  frontend: {frontend["type"]} ({connection})')
            try:
                if frontend['type'] == 'status':
                    t = StatusFrontendThread(frontend, nthreads)
                elif frontend['type'] == 'http-api':
                    t = HttpApiFrontendThread(frontend, nthreads)
                elif frontend['type'] == 'tcp':
                    t = TcpFrontendThread(frontend, nthreads)
                elif frontend['type'] == 'mqtt':
                    t = MqttFrontendThread(frontend, nthreads)
                elif frontend['type'] == 'midi':
                    t = MidiFrontendThread(frontend, nthreads)
                else:
                    logging.error(f'  Unknown frontend type "{frontend["type"]}"')
                    continue
                t.daemon = True
                threads.append(t)
                nthreads['frontend'][t.name] = t
                t.start()
            except RecoverableError as e:
                logging.error(f'  Could not initialize the "{frontend["type"]}" frontend. {e}')

    while True:
        time.sleep(1)


def main():
    parser = argparse.ArgumentParser("OpenSwitcher proxy")
    parser.add_argument("--config", help="Config file to use", default="/etc/openswitcher/proxy.toml")
    args = parser.parse_args()

    run(args.config)


if __name__ == '__main__':
    main()



import socketserver
import logging
import time


class VirtualHyperdeck(socketserver.BaseRequestHandler):
    def setup(self):
        self.remote = True
        self.uid = '1234'

        self.notify_transport = False
        self.notify_slot = False
        self.notify_remote = False
        self.notify_configuration = False

        self.transport_status = 'stopped'
        self.transport_speed = 0
        self.transport_slot_id = 1
        self.timecode = "01:00:00:00"
        self.display_timecode = "01:00:00:00"
        self.clip_id = 0
        self.video_format = "720p50"
        self.loop = False

        self.clips = [
            (1, "Test clip", 30),
            (2, "Another clip", 45)
        ]

    def handle(self):
        logging.info('connected')
        self.send(b'500 connection info:\r\nprotocol version: 1.9\r\nmodel: Linux\r\n\r\n')
        while True:
            data = self.request.recv(1024)
            if len(data) == 0:
                logging.info('disconnected')
                break
            logging.debug('< ' + data.decode())
            self.handle_line(data.decode().strip())

    def send(self, data):
        logging.debug('> ' + data.decode())
        self.request.sendall(data)

    def parse_line(self, line):
        parameters = {}
        parts = line.split(' ')
        name = parts[0].rstrip(':')
        key = None
        for part in parts[1:]:
            if part[-1] == ':':
                key = part.rstrip(':')
            else:
                if key not in parameters:
                    parameters[key] = ''
                parameters[key] += part
                parameters[key] = parameters[key].strip()
        return name, parameters

    def handle_line(self, line):
        if line.startswith('notify:'):
            _, parameters = self.parse_line(line)
            if 'transport' in parameters:
                self.notify_transport = parameters['transport'] == 'true'
            if 'slot' in parameters:
                self.notify_slot = parameters['slot'] == 'true'
            if 'remote' in parameters:
                self.notify_remote = parameters['remote'] == 'true'
            if 'configuration' in parameters:
                self.notify_configuration = parameters['configuration'] == 'true'

            self._send_notify(sync=True)

        elif line.startswith('transport info'):
            self._send_transport_info(sync=True)
        elif line == 'remote':
            self._send_remote_info(sync=True)
        elif line == 'clips count':
            self._send_clips_count()
        elif line.startswith('slot info'):
            _, parameters = self.parse_line(line)
            self._send_slot_info(sync=True)
        elif line.startswith('clips get: clip'):
            _, parameters = self.parse_line(line)
            self._send_clip_info(sync=True, clip=int(parameters['id']))
        elif line.startswith('goto: clip'):
            _, parameters = self.parse_line(line)
            self.clip_id = int(parameters['id'])
            self.send(b'200 ok\r\n')
            if self.notify_transport:
                self._send_transport_info(sync=False)
        else:
            time.sleep(0.5)

    def _send_notify(self, sync=True):
        if sync:
            response = b'209 notify:\r\n'
        else:
            response = b'509 remote info:\r\n'
        response += 'transport: {}\r\n'.format('true' if self.notify_transport else 'false').encode()
        response += 'slot: {}\r\n'.format('true' if self.notify_slot else 'false').encode()
        response += 'remote: {}\r\n'.format('true' if self.notify_remote else 'false').encode()
        response += 'configuration: {}\r\n'.format('true' if self.notify_configuration else 'false').encode()
        response += b'\r\n'
        self.send(response)

    def _send_remote_info(self, sync=True):
        if sync:
            response = b'210 remote info:\r\n'
        else:
            response = b'510 remote info:\r\n'
        response += b'enabled: true\r\n'
        response += b'override: true\r\n'
        response += b'\r\n'
        self.send(response)

    def _send_transport_info(self, sync=True):
        if sync:
            response = b'208 transport info:\r\n'
        else:
            response = b'508 transport info:\r\n'

        response += 'status: {}\r\n'.format(self.transport_status).encode()
        response += 'speed: {}\r\n'.format(self.transport_speed).encode()
        response += 'slot id: {}\r\n'.format(self.transport_slot_id or 'none').encode()
        response += 'timecode: {}\r\n'.format(self.timecode).encode()
        response += 'display timecode: {}\r\n'.format(self.display_timecode).encode()
        response += 'clip id: {}\r\n'.format(self.clip_id).encode()
        response += 'video format: {}\r\n'.format(self.video_format or 'none').encode()
        response += 'loop: {}\r\n'.format('true' if self.loop else 'false').encode()
        response += b'\r\n'
        self.send(response)

    def _send_slot_info(self, sync=True):
        if sync:
            response = b'202 slot info:\r\n'
        else:
            response = b'502 slot info:\r\n'
        response += b'slot id: 1\r\n'
        response += b'status: mounted\r\n'
        response += b'volume name: SD\r\n'
        response += b'recording time: 0\r\n'
        response += 'video format: {}\r\n'.format(self.video_format).encode()
        response += b'\r\n'
        self.send(response)

    def _send_configuration(self, sync=True):
        if sync:
            response = b'211 configuration:\r\n'
        else:
            response = b'511 configuration:\r\n'
        response += b'audio input: embedded\r\n'
        response += b'video input: HDMI\r\n'
        response += b'file format: H.264High\r\n'
        response += b'\r\n'
        self.send(response)

    def _send_clips_count(self):
        response = b'214 clips count:\r\n'
        response += 'clip count: {}\r\n'.format(len(self.clips)).encode()
        response += b'\r\n'
        self.send(response)

    def _send_clip_info(self, sync=True, clip=0):
        response = b'205 clips info:\r\n'
        response += 'clip count: {}\r\n'.format(len(self.clips)).encode()
        for clip in self.clips:
            cliplen = '00:00:{}:00'.format(clip[2])
            response += '{}: {} 00:00:00:00 {}\r\n'.format(clip[0], clip[1], cliplen).encode()
        response += b'\r\n'
        self.send(response)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    server = socketserver.TCPServer(('0.0.0.0', 9993), VirtualHyperdeck, bind_and_activate=False)
    server.allow_reuse_address = True
    server.server_bind()
    server.server_activate()
    server.serve_forever()



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import logging
import time
from urllib.parse import urlparse, quote

import gi

from gtk_switcher.preferences import PreferencesWindow
from pyatem import locate

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, GdkPixbuf

gi.require_version('Handy', '1')
from gi.repository import Handy


class ConnectionWindow:
    def __init__(self, parent, connection, application):
        builder = Gtk.Builder()
        builder.set_translation_domain("openswitcher")
        builder.add_from_resource('/nl/brixit/switcher/ui/connection.glade')
        builder.connect_signals(Handler(builder, parent, application, connection))

        window = builder.get_object("preferences_window")

        window.set_transient_for(parent)
        window.set_modal(True)

        window.show_all()


class Item(GObject.GObject):
    text = GObject.property(type=str)

    def __init__(self, text):
        GObject.GObject.__init__(self)
        self.text = text


class Handler:
    def __init__(self, builder, parent, application, connection):
        self.builder = builder
        self.parent = parent
        self.application = application
        self.connection = connection
        self.window = builder.get_object('preferences_window')

        self.log = logging.getLogger('ConnectionWindow')

        self.ipaddress = builder.get_object('ipaddress')
        self.username = builder.get_object('username')
        self.password = builder.get_object('password')
        self.device = builder.get_object('device')
        self.avahi_list = builder.get_object('avahi_list')
        self.preferences_button = builder.get_object('preferences')

        self.connection_udp = builder.get_object('connection_udp')
        self.connection_usb = builder.get_object('connection_usb')
        self.connection_tcp = builder.get_object('connection_tcp')

        self.preferences_button.set_sensitive(self.connection.connected)

        self.settings = Gio.Settings.new('nl.brixit.Switcher')
        self.settings.connect('changed::switcher-ip', self.on_switcher_ip_changed)

        self.on_switcher_ip_changed(self.settings, 'switcher-ip')
        locate.listen(self.on_discovered)

    def on_discovered(self, name, subtitle, proto, address):
        GLib.idle_add(self.on_discovered_ui, name, subtitle, proto, address)

    def on_discovered_ui(self, name, subtitle, proto, address):
        # Drop placeholder since it messes with libhandy css
        self.avahi_list.set_placeholder(None)

        label = Gtk.Label(address[0])
        label.address = address
        label.proto = proto
        ar = Handy.ActionRow()
        ar.set_title(name)
        ar.set_subtitle(subtitle)
        ar.add(label)
        time.sleep(0.1)
        self.avahi_list.add(ar)
        time.sleep(0.1)
        self.avahi_list.show_all()

    def on_lost(self, address):
        self.log.info('lost', address)

    def on_avahi_activate(self, widget, row):
        label = list(row)[0]
        if label.proto == 'udp':
            self.connection_udp.set_active(True)
            self.ipaddress.set_text(str(label.address[0]))
        elif label.proto == 'tcp':
            self.connection_tcp.set_active(True)
            self.ipaddress.set_text(f'{label.address[0]}:{label.address[1]}')
        elif label.proto == 'usb':
            self.ipaddress.set_text('')
            self.connection_usb.set_active(True)

    def on_destroy(self, *args):
        locate.stop()

    def on_save_clicked(self, *args):
        if self.connection_udp.get_active():
            self.settings.set_string('switcher-ip', self.ipaddress.get_text())
        elif self.connection_usb.get_active():
            self.settings.set_string('switcher-ip', "0.0.0.0")
        else:
            url = 'tcp://'
            username = self.username.get_text()
            password = self.password.get_text()
            if username != "":
                url += f"{quote(username)}:{quote(password)}@"
            url += self.ipaddress.get_text()
            url += '/' + self.device.get_text()
            self.log.info("New connection url: " + str(url))
            self.settings.set_string('switcher-ip', url)
        self.window.close()

    def fix_widget_state(self):
        if self.connection_udp.get_active():
            self.ipaddress.set_sensitive(True)
            self.username.set_sensitive(False)
            self.password.set_sensitive(False)
            self.device.set_sensitive(False)
        elif self.connection_usb.get_active():
            self.ipaddress.set_sensitive(False)
            self.username.set_sensitive(False)
            self.password.set_sensitive(False)
            self.device.set_sensitive(False)
        else:
            self.ipaddress.set_sensitive(True)
            self.username.set_sensitive(True)
            self.password.set_sensitive(True)
            self.device.set_sensitive(True)

    def on_connection_change(self, *args):
        self.fix_widget_state()

    def on_switcher_ip_changed(self, settings, key, *args):
        if key == 'switcher-ip':
            value = self.settings.get_string('switcher-ip')
            if value.startswith('tcp://'):
                part = urlparse(value)
                address = part.hostname
                if part.port:
                    address += ':' + str(part.port)
                self.ipaddress.set_text(address)

                if part.username is None:
                    self.username.set_text("")
                else:
                    self.username.set_text(part.username)

                if part.password is None:
                    self.password.set_text("")
                else:
                    self.password.set_text(part.password)
                self.device.set_text(part.path[1:])

                self.connection_tcp.set_active(True)

            elif value == '0.0.0.0':
                self.connection_usb.set_active(True)
                self.ipaddress.set_text("")
            else:
                self.connection_udp.set_active(True)
                self.ipaddress.set_text(value)
            self.fix_widget_state()

    def on_preferences_clicked(self, *args):
        PreferencesWindow(self.parent, self.application, self.connection)
